# -*- coding: utf-8 -*-
# Module name: msgbus.py
# Module Description: Python wrapper for msgbus
#
# Copyright © 2008 MWneo Corporation
#
# $Id: msgbus.py 14404 2014-04-08 11:53:45Z rrocha $
# $Revision: 14404 $
# $Date: 2014-04-08 08:53:45 -0300 (Tue, 08 Apr 2014) $
# $Author: rrocha $

"""
Python wrapper for native message-bus library.

This module wraps all native message-bus functions with a object-oriented API.
All "release" related functions are not wrapped because this module automatically
handles the native releasing thru python garbage collector.

When this module is imported, the native "msgbus" library will be loaded, so it is
possible that an exception is raised during the module initialization (if any problem
happen loading the native library).

Minimal "Hello world" example:
>>> ctx = MBEasyContext("python test")
>>> while True:
        msg = ctx.MB_EasyGetMessage()
        if not msg:
            break
        print "Received message. Token: %d. Format: %d. Data: %s."%(msg.token,msg.format,msg.data)
        ctx.MB_EasyReplyMessage(msg)
>>> print "Done!"

"""

import httplib
import socket
import traceback
import ctypes
import ctypes.util
import os.path
import base64
import sys
from ctypes import c_int, c_void_p, c_char_p, c_int32, c_uint32, c_int64, Structure, POINTER, string_at, cdll
from syserrors import SE_SUCCESS, SE_TIMEOUT, SE_NOTFOUND, SE_UNDEFINED
from systools import sys_errget

_dll = None


def _find_library(name):
    """ finds a native library in the system, given it's name """
    platform = "win32" if sys.platform.startswith("win") else "linux" if (
        sys.platform.find("linux") != -1) else "macos"
    prefix, suffix, envvar = ({"win32": ("", ".dll", "PATH"), "linux": (
        "lib", ".so", "LD_LIBRARY_PATH"), "macos": ("lib", ".dylib", "DYLD_LIBRARY_PATH")})[platform]
    filename = "%s%s%s" % (prefix, name, suffix)
    for libpath in (os.path.join(dirname, filename) for dirname in ["./"] + os.environ.get(envvar, "").split(os.path.pathsep)):
        if os.path.exists(libpath):
            return libpath
    libpath = ctypes.util.find_library(name)
    if libpath and os.path.exists(libpath):
        return libpath
    return None


# Startup flags
SU_FRESH = 0x00000001

# START OF TOKEN LIST
TK_SYS_ACK = 0xF0100001
TK_SYS_NAK = 0xF0100002
TK_SYS_CHECKFORUPDATES = 0x00100003
TK_HV_COMPSTART = 0x00200001
TK_HV_COMPSTOP = 0x00200002
TK_HV_COMPKILL = 0xF0200002
TK_HV_COMPLIST = 0x00200003
TK_HV_QUERY_SERVICE = 0x00200004
TK_HV_FIND_SERVICE = 0x00200005
TK_HV_QUERY_BY_TYPE = 0x00200006
TK_HV_START_COMP = 0x00200007
TK_HV_SERVICEONLINE = 0x00200008
TK_HV_SERVICEOFFLINE = 0x00200009
TK_HV_RESTART = 0x0020000A
TK_HV_GLOBALRESTART = 0x0020000B
TK_HV_GETSTATE = 0x0020000C
TK_HV_SETSTATE = 0x0020000D
TK_HV_LOG = 0x0020000E
TK_HV_GETLICENSE = 0x0020000F
TK_HV_LOGLEVEL = 0xF0200010
TK_HV_LISTHVS = 0x00200020
TK_HV_CHECK_LICENSE_SIGNATURE = 0x00200030
TK_CMP_TERM_NOW = 0xF0300001
TK_CMP_GETSERVICES = 0xF0300002
TK_CMP_GETDEPENDS = 0xF0300003
TK_CMP_REGSERVICE = 0xF0300004
TK_CMP_START = 0xF0300005
TK_CMP_MESSAGE = 0xF0300006
TK_EVT_EVENT = 0xF0400001
TK_EVT_SUBSCRIBE = 0xF0400002
TK_EVT_UNSUBSCRIBE = 0xF0400003
TK_EVT_RUNTESTS = 0xF0400004
TK_EVT_GETLISTENERS = 0xF0400005
TK_POS_GETMODEL = 0xF0500001
TK_POS_SETSCREEN = 0xF0500002
TK_POS_EXECUTEACTION = 0xF0500003
TK_POS_MODELSETCUSTOM = 0xF0500004
TK_POS_SETINFOMSG = 0xF0500005
TK_POS_SHOWDIALOG = 0xF0500006
TK_POS_DIALOGRESP = 0xF0500007
TK_POS_BUSINESSBEGIN = 0xF0500008
TK_POS_BUSINESSEND = 0xF0500009
TK_POS_USERLOGIN = 0xF050000A
TK_POS_USERPAUSE = 0xF050000B
TK_POS_USERLOGOUT = 0xF050000C
TK_POS_SETDEFSERVICE = 0xF050000D
TK_POS_SETLANG = 0xF050000E
TK_POS_SETMAINSCREEN = 0xF050000F
TK_POS_SETPOD = 0xF0500010
TK_POS_SETFUNCTION = 0xF0500011
TK_POS_USEROPEN = 0xF0500012
TK_POS_SETMODIFIERSDATA = 0xF0500013
TK_POS_ASYNCACTION = 0xF0500014
TK_POS_GETSTATE = 0xF0500015
TK_POS_GETCONFIG = 0xF0500016
TK_POS_LISTUSERS = 0xF0500017
TK_POS_SETRECALLDATA = 0xF0500018
TK_POS_GETPOSLIST = 0xF0500019
TK_POS_UPDATEDIALOG = 0xF050001A
TK_KDS_GETMODEL = 0xF0B00001
TK_KDS_UPDATEVIEW = 0xF0B00002
TK_KDS_REFRESH = 0xF0B00003
TK_KDS_REFRESH_END = 0xF0B00004
TK_KDS_SET_PROD_STATE = 0xF0B00005
TK_KDS_UNDO_SERVE = 0xF0B00006
TK_KDS_CHANGE_VIEW = 0xF0B00007
TK_KDS_TOGGLE_TAG_LINE = 0xF0B00008
TK_KDS_GLOBAL_SERVE = 0xF0B00009
TK_KDS_REFRESH_VIEW = 0xF0B00010
TK_KDS_DISABLE_VIEW = 0xF0B00011
TK_KDS_UPDATE_INFO = 0xF0B00012
TK_KDS_ENABLE_VIEW = 0xF0B00013
TK_KDS_RELOAD = 0xF0B00014
TK_KIOSK_GETMODEL = 0xF1D00001
TK_KIOSK_SETLANG = 0xF1D00002
TK_KIOSK_GETSTATE = 0xF1D00003
TK_KIOSK_SETSTATE = 0xF1D00004
TK_KIOSK_CHECKFORUPDATES = 0xF1D01001
TK_KIOSK_EXECUTEACTION = 0xF1D01002
TK_KIOSK_PROVIDER_RESERVE = 0xF1D02001
TK_KIOSK_PROVIDER_CONFIRM = 0xF1D02002
TK_KIOSK_PROVIDER_DISMISS = 0xF1D02003
TK_KIOSK_TRANSACTION_PICTURE = 0xF1D03001
TK_FISCAL_CMD = 0xF0600001
TK_PRN_PRINT = 0xF0600101
TK_PRN_GETSTATUS = 0xF0600102
TK_CDRAWER_OPEN = 0xF0600201
TK_CDRAWER_GETSTATUS = 0xF0600202
TK_LDISP_WRITE = 0xF0600301
TK_LDISP_MARQUEE = 0xF0600302
TK_MSR_REQUESTSWIPE = 0xF0600402
TK_COIN_DISPENSE = 0xF0600801
TK_COIN_GETSTATUS = 0xF0600802
TK_COIN_SETUP = 0xF0600803
TK_COIN_RESET = 0xF0600804
TK_COIN_EXAMINE = 0xF0600805
TK_COIN_ADJUST = 0xF0600806
TK_COIN_COLLECT = 0xF0600807
TK_IMGCAP_CAPTURE = 0xF0600901
TK_BUZZER_SOUND = 0xF0600B01
TK_CTRLUNIT_REGISTER_SALE = 0xF0600B01
TK_CTRLUNIT_REGISTER_RETURN = 0xF0600B02
TK_CTRLUNIT_WRITE_JOURNAL = 0xF0600B03
TK_DRV_IOCTL = 0xF060FFFF
TK_I18N_GETTABLE = 0x00700001
TK_I18N_GETMSG = 0x00700002
TK_I18N_TRANSLATE = 0x00700003
TK_I18N_GETLANGS = 0x00700004
TK_I18N_GETTABLEJSON = 0x00700005
TK_L10N_TOCURRENCY = 0x00700006
TK_L10N_TONUMBER = 0x00700007
TK_L10N_TODATE = 0x00700008
TK_L10N_TOTIME = 0x00700009
TK_L10N_TODATETIME = 0x0070000A
TK_L10N_FROMCURRENCY = 0x0070000B
TK_L10N_FROMNUMBER = 0x0070000C
TK_L10N_FROMDATE = 0x0070000D
TK_L10N_FROMTIME = 0x0070000E
TK_L10N_FROMDATETIME = 0x0070000F
TK_PERSIST_QUERYEXEC = 0x00800001
TK_PERSIST_CONNALLOC = 0x00800002
TK_PERSIST_CONNRELEASE = 0x00800003
TK_PERSIST_SCRIPTEXEC = 0x00800004
TK_PERSIST_PROCEXEC = 0x00800005
TK_PERSIST_EXPORTDB = 0x00800006
TK_PERSIST_IMPORTDB = 0x00800007
TK_PERSIST_LISTDBS = 0x00800008
TK_PERSIST_GETINSTNBR = 0x00800009
TK_PURGE_DB = 0x0080000A
TK_PERSIST_INITTHRDTERM = 0x0080000B
TK_PERSIST_INITINPROGRESS = 0x0080000C
TK_PERSIST_SEQUENCER_INCREMENT = 0x0080000D
TK_PERSIST_GET_PROCEDURES_LIST = 0x0080000E
TK_PERSIST_GET_PROCEDURE_PARAMS = 0x0080000F
TK_PERSIST_SEQUENCER_RESET = 0x00800010
TK_PERSIST_CHECKCONNECTION = 0x00800011
TK_REPLICATE = 0xF0800008
TK_REPLICA_ACQUIRE = 0xF0800009
TK_REPLICA_LISTDBS = 0xF0800010
TK_REPLICA_GETCHKPNT = 0xF0800011
TK_REPLICA_GETFAILOVERCHKPNT = 0xF0800012
TK_REPLICA_GETUSERVERSION = 0xF0800013
TK_SLCTRL_PMGR_MAINMENU = 0x00900001
TK_SLCTRL_PMGR_CUSTOMPARAM = 0x00900002
TK_SLCTRL_PMGR_PRODUCTSBYTAGS = 0x00900003
TK_SLCTRL_PMGR_PRODUCTDIM = 0x00900004
TK_SLCTRL_OMGR_DOSALE = 0x00A00001
TK_SLCTRL_OMGR_ORDERPICT = 0x00A00002
TK_SLCTRL_OMGR_VOIDLINE = 0x00A00003
TK_SLCTRL_OMGR_VOIDORDER = 0x00A00004
TK_SLCTRL_OMGR_LISTOPTIONS = 0x00A00005
TK_SLCTRL_OMGR_ORDERTOTAL = 0x00A00006
TK_SLCTRL_OMGR_CLEAROPTION = 0x00A00007
TK_SLCTRL_OMGR_SPLITORDER = 0x00A00008
TK_SLCTRL_OMGR_SPLITORDERLINE = 0x00A00009
TK_SLCTRL_OMGR_DOOPTION = 0x00A0000A
TK_SLCTRL_OMGR_STOREORDER = 0x00A0000B
TK_SLCTRL_OMGR_RECALLORDER = 0x00A0000C
TK_SLCTRL_OMGR_LISTORDERS = 0x00A0000D
TK_SLCTRL_OMGR_CHANGEQTY = 0x00A0000E
TK_SLCTRL_OMGR_ORDERREOPEN = 0x00A0000F
TK_SLCTRL_OMGR_DOTENDER = 0x00A00010
TK_SLCTRL_OMGR_CLEARTENDERS = 0x00A00011
TK_SLCTRL_OMGR_LISTMODIFIERS = 0x00A00012
TK_SLCTRL_OMGR_CHANGEMODIFIER = 0x00A00013
TK_SLCTRL_OMGR_CHANGEDIMENSION = 0x00A00014
TK_SLCTRL_OMGR_RESETCURRORDER = 0x00A00015
TK_SLCTRL_OMGR_SETLSTSALELINE = 0x00A00016
TK_SLCTRL_OMGR_UPDPROPERTIES = 0x00A00017
TK_SLCTRL_OMGR_CREATEORDER = 0x00A00018
TK_SLCTRL_OMGR_CHANGEMODIFIERS = 0x00A00019
TK_SLCTRL_OMGR_VOIDHISTORY = 0x00A0001A
TK_SLCTRL_OMGR_PRICEOVERWRITE = 0x00A0001B
TK_SLCTRL_OMGR_LISTOPTSOLUTIONS = 0x00A0001C
TK_SLCTRL_OMGR_SETCUSTOMPROPERTY = 0x00A0001D
TK_SLCTRL_OMGR_GETCUSTOMPROPERTY = 0x00A0001E
TK_SLCTRL_OMGR_LIST_DISCOUNTS = 0x00A0001F
TK_SLCTRL_OMGR_APPLY_DISCOUNT = 0x00A00020
TK_SLCTRL_OMGR_CLEAR_DISCOUNT = 0x00A00021
TK_SLCTRL_OMGR_MERGE_ORDERS = 0x00A00022
TK_SLCTRL_OMGR_EXTRAMODIFIERS = 0x00A00023
TK_SLCTRL_OMGR_DOMULTIOPTIONS = 0x00A00024
TK_SLCTRL_OMGR_ADDCOMMENT = 0x00A00025
TK_SLCTRL_OMGR_DELCOMMENT = 0x00A00026
TK_SLCTRL_OMGR_UPDCOMMENT = 0x00A00027
TK_SLCTRL_OMGR_SETCUSTOMPROPERTIES = 0x00A00028
TK_SLCTRL_OMGR_FRACTIONATELINE = 0x00A00029
TK_SLCTRL_OMGR_REVERTLINEFRACTION = 0x00A00030
TK_SLCTRL_RESET = 0x02600001
TK_PROD_REFRESHVIEW = 0x00C00001
TK_PROD_SETORDERSTATE = 0x00C00002
TK_PROD_CRASHPOINT = 0x00C00003
TK_PROD_CAPTUREIMAGE = 0x00C00004
TK_PROD_RETRIEVEIMAGE = 0x00C00005
TK_PROD_SIDEONOFF = 0x00C00006
TK_PROD_UNDOSERVE = 0x00C00007
TK_PROD_TOGGLETAGLINE = 0x00C00008
TK_PROD_GLOBAL_SERVE = 0x00C00009
TK_PROD_DISABLE_VIEW = 0x00C00010
TK_PROD_ENABLE_VIEW = 0x00C00011
TK_PROD_REDIRECT_BOX = 0x00C00012
TK_PROD_RESTORE_BOX = 0x00C00013
TK_REPORT_GENERATE = 0x00D00001
TK_ACCOUNT_TRANSFER = 0x00E00001
TK_ACCOUNT_GETTRANSFERS = 0x00E00002
TK_ACCOUNT_GETTOTALS = 0x00E00003
TK_ACCOUNT_GIFTCARDACTIVITY = 0x00E00004
TK_ACCOUNT_EFTACTIVITY = 0x00E00005
TK_EFT_REQUEST = 0x00F00001
TK_GENESIS_GETVERSION = 0x01000001
TK_GENESIS_LISTFILES = 0x01000002
TK_GENESIS_GETFILE = 0x01000003
TK_GENESIS_LISTPKGS = 0x01000004
TK_GENESIS_APPLYUPDATE = 0x01000005
TK_GENESIS_ROLLBACKUPDATE = 0x01000006
TK_TIMEPUNCH_OPER = 0x01100001
TK_TIMEPUNCH_REPORT = 0x01100002
TK_TIMEPUNCH_GETJOBCODES = 0x01100003
TK_USERCTRL_AUTHENTICATE = 0x01200001
TK_USERCTRL_GETINFO = 0x01200002
TK_USERCTRL_GETINFO_BADGE = 0x01200003
TK_STORECFG_GET = 0x01300001
TK_STORECFG_GETFULL = 0x01300002
TK_STORECFG_SET = 0x01300003
TK_AUDITLOG_RESYNC = 0xF1400001
TK_BOINVENTORY_UNUSED = 0x01500000
TK_BOINVENTORY_MANAGE_SUPPLIER = 0x01500001
TK_BOINVENTORY_MANAGE_SUPPLIER_PRODUCTS = 0x01500002
TK_BOINVENTORY_MANAGE_SUPPLIER_INVOICES = 0x01500003
TK_BOINVENTORY_MANAGE_STORAGE_AREAS = 0x01500004
TK_BOINVENTORY_MANAGE_DEPARTMENTS = 0x01500005
TK_BOINVENTORY_MANAGE_GROUPS = 0x01500006
TK_BOINVENTORY_MANAGE_ITEMS = 0x01500007
TK_BOINVENTORY_MANAGE_BOM = 0x01500008
TK_BOINVENTORY_MANAGE_INVENTORY_COUNTS = 0x01500009
TK_BOINVENTORY_IMPORT_POS_SALES = 0x0150000A
TK_BOINVENTORY_INVENTORY_CONTROL = 0x0150000B
TK_BOINVENTORY_INVOICES_AVAILABLE = 0x0150000C
TK_BOINVENTORY_SALES_AVAILABLE = 0x0150000D
TK_BOINVENTORY_INVENTORY_TRANSACTION = 0x0150000E
TK_BOINVENTORY_DELIVERY_TRANSACTION = 0x0150000F
TK_BOINVENTORY_POS_SALES_TRANSACTION = 0x01500010
TK_BOINVENTORY_DELIVERY_TRANSACTION_DETAILS = 0x01500011
TK_BOINVENTORY_POS_SALES_TRANSACTION_DETAILS = 0x01500012
TK_BOINVENTORY_COUNT_ADJUSTMENT_TRANSACTION = 0x01500013
TK_BOINVENTORY_DELIVERY_ADJUSTMENT_TRANSACTION = 0x01500014
TK_BOINVENTORY_SUBINVENTORY_TRANSFER_TRANSACTION = 0x01500015
TK_BOINVENTORY_INBOUND_TRANSFER_TRANSACTION = 0x01500016
TK_BOINVENTORY_OUTBOUND_TRANSFER_TRANSACTION = 0x01500017
TK_BOINVENTORY_SUPPLIER_RETURN_TRANSACTION = 0x01500018
TK_BOINVENTORY_WIP_SCRAP_TRANSACTION = 0x01500019
TK_BOINVENTORY_INVENTORY_RESET = 0x0150001A
TK_BOINVENTORY_REPORTS = 0x0150001B
TK_BOINVENTORY_MANAGE_ASSEMBLY = 0x0150001C
TK_BOINVENTORY_MANAGE_ITEM_PACKAGES = 0x0150001D
TK_BOINVENTORY_MANAGE_PLU_LIST = 0x0150001E
TK_BOCASH_REPORTS = 0x01600000
TK_BOCASH_MANAGE_STORECFG = 0x01600001
TK_BOCASH_MANAGE_BUSINESSWEEK = 0x01600002
TK_BOCASH_MANAGE_DRAWERCOUNT_TEMPLATE = 0x01600003
TK_BOCASH_MANAGE_BUSINESSDATE = 0x01600004
TK_BOCASH_POSEVENTS_DAYOPEN = 0x01600005
TK_BOCASH_POSEVENTS_DAYCLOSE = 0x01600006
TK_BOCASH_POSEVENTS_SKIM = 0x01600007
TK_BOCASH_POSEVENTS_DRAWERCHANGE = 0x01600008
TK_BOCASH_MANAGE_DRAWERCOUNT = 0x01600009
TK_BOCASH_MANAGE_DRAWERCOUNT_DETAILS = 0x0160000A
TK_BOCASH_MANAGE_TRANSACTIONS = 0x0160000B
TK_BOCASH_MANAGE_TRANSACTION_DETAILS = 0x0160000C
TK_BOCASH_POSEVENTS_HISTORY = 0x0160000D
TK_BOCASH_FORCE_WEEK_CLOSE = 0x0160000E
TK_BOREPORT_REPORTS = 0x01800000
TK_BOREPORT_MANAGE_TEMPLATE = 0x01800001
TK_BOREPORT_MANAGE_WORKBOOK = 0x01800002
TK_BOREPORT_MANAGE_WORKSHEET = 0x01800003
TK_BOREPORT_MANAGE_WORKSHEET_CELLS = 0x01800004
TK_BOREPORT_WORKBOOK_FILE = 0x01800005
TK_BOREPORT_WORKBOOK_SOLVER = 0x01800006
TK_BOREPORT_WORKBOOK_PREVIEW = 0x01800007
TK_BOREPORT_GETCONFIG = 0x01800008
TK_BOACCOUNT_REPORTS = 0x01700000
TK_BOACCOUNT_MANAGE_TRNTYPE = 0x01700001
TK_BOACCOUNT_MANAGE_CHARTOFACCOUNTS = 0x01700002
TK_BOACCOUNT_MANAGE_JOURNALTEMPLATE = 0x01700003
TK_BOACCOUNT_MANAGE_JOURNAL = 0x01700004
TK_BOACCOUNT_MANAGE_TRIALBALANCE = 0x01700005
TK_BOLABOR_REPORTS = 0x01C00000
TK_BOLABOR_MANAGE_EMPLOYEES = 0x01C00001
TK_BOLABOR_MANAGE_TIMEPUNCHES = 0x01C00002
TK_BOLABOR_MANAGE_EMPLOYEE_CUSTOMDATA = 0x01C00003
TK_CASHLESS_REPORT = 0x01900001
TK_GUI_REQUESTFOCUS = 0x01A00001
TK_GUI_SETURL = 0x01A00002
TK_GUI_EVALJS = 0x01A00003
TK_GUI_MINIMIZE = 0x01A00004
TK_CALLCENTER_GETSTATUS = 0x01B00001
TK_CALLCENTER_AGTLOGIN = 0x01B00002
TK_CALLCENTER_SETAGTSTATE = 0x01B00003
TK_CALLCENTER_NEWCALL = 0x01B00004
TK_CALLCENTER_CALLENDED = 0x01B00005
TK_CALLCENTER_AGTSTATECHANGED = 0x01B00006
TK_CALLCENTER_RECONNECT = 0x01B00007
TK_CALLCENTER_SETVOLUME = 0x01B00008
TK_CALLCENTER_TRANSFER = 0x01B00009
TK_CALLCENTER_HEARTHBEAT = 0x01B0000A
TK_CALLCENTER_EXITAPP = 0x01B0000B
TK_CALLCENTER_SETFAILOVER = 0x01B0000C
TK_CALLCENTER_SHOWALERT = 0x01B0000D
TK_CALLCENTER_MANAGEBTREST = 0x01B0000E
TK_LS_RETRIEVE_CACHE = 0x01E00001
TK_LS_STORE_CACHE = 0x01E00002
TK_LS_CREATE_SESSION = 0x01E00003
TK_LS_STORE_PURCHASE = 0x01E00004
TK_LS_VOID_PURCHASE = 0x01E00005
TK_BUF_BUFFER_PURCHASE = 0x01F00001
TK_BUF_PAY_ORDER = 0x01F00002
TK_BUF_STORE_API_CALL = 0x01F00003
TK_BUF_FINISH_ORDER = 0x01F00004
TK_ML_REFRESH_CACHE = 0x02000001
TK_KIOSKCTRL_DBUPDATE = 0x02100001
TK_KIOSKCTRL_GENESISUPDATE = 0x02100002
TK_KIOSKCTRL_IMAGESUPDATE = 0x02100003
TK_CATERING_NEWEVENT = 0x02200001
TK_CATERING_STORECUSTOMER = 0x02200002
TK_CATERING_SEARCHCUSTOMER = 0x02200003
TK_CATERING_UPDATECUSTOMER = 0x02200004
TK_CATERING_UPDATEEVENT = 0x02200005
TK_CATERING_UPDATEBILLINGID = 0x02200006
TK_CATERING_ORDERPICT_TO_JSON = 0x02200007
TK_CATERING_GETCAFEINFO = 0x02200008
TK_CATERING_GETEVENT = 0x02200009
TK_ROPECTRL_SEND_XML = 0x02300001
TK_ROPECTRL_ROPE_XML = 0x02300002
TK_BOH_MESSAGE = 0x02400001
TK_BOH_UPLOADORDER = 0x02400002
TK_LOYALTY_AUTHENTICATE = 0x02500001
TK_LOYALTY_BALANCE = 0x02500002
TK_LOYALTY_APPLY = 0x02500003
TK_LOYALTY_VOID = 0x02500004
TK_LOYALTY_REFUND = 0x02500005
TK_LOYALTY_COMMAND = 0x02500006
# END OF TOKEN LIST

# Data formats
FM_UNDEFINED = 0x00000000
FM_XML = 0x00000001
FM_PARAM = 0x00000002
FM_ADDR_PORT = 0x00000003
FM_INT32 = 0x00000004
FM_INT64 = 0x00000005
FM_STRING = 0x00000006
FM_STR_ARRAY_PIPE = 0x00000007
FM_IMG_JPG = 0x00000008
FM_IMG_PNG = 0x00000009
FM_IMG_GIF = 0x0000000A
FM_PDF = 0x0000000B
FM_SWAGGER = 0x0000000C

#
# Structure definitions
#


class st_msg(Structure):
    _fields_ = [("sign", c_int32),       # signature
                ("size", c_int32),       # data size
                ("token", c_uint32),     # token
                ("format", c_int32),     # format
                ("destname", c_char_p),  # destination service name (optional)
                ("data", c_void_p),      # data
                ("trailer", c_int32),    # trailer
                ("crc", c_int32),        # crc32
                ("remotehost", c_char_p)]  # remote host name or IP address - for incoming messages only


st_msg_p = POINTER(st_msg)


class st_service(Structure):
    _fields_ = [("addr", c_char_p),      # ip addr
                ("port", c_int32),       # port
                ("name", c_char_p), ]    # service name


st_service_p = POINTER(st_service)


class st_easymb(Structure):
    _fields_ = [("pctx", c_void_p),      # a plain message bus context created by MB_EasyInitialize
                ("ser", st_service_p),   # service information
                ("hv", st_service_p),    # hypervisor information
                ("opaque", c_void_p)]    # reserved for internal use


st_easymb_p = POINTER(st_easymb)

c_enun = c_int
c_enun_p = POINTER(c_enun)
c_int_p = POINTER(c_int)
c_int32_p = POINTER(c_int32)
c_char_pp = POINTER(c_char_p)

#
# DLL Function definitions
#

_funcdefs = (
    ('MB_CreateService', st_service_p, (c_char_p, c_int32, c_void_p, c_char_p, c_int32)),
    ('MB_Initialize', c_void_p, (st_service_p, c_char_p, c_enun_p, c_char_p, c_int32)),
    ('MB_GetMessage', st_msg_p, (c_void_p, c_enun, c_enun, c_char_p, c_int32)),
    ('MB_SendMessage', st_msg_p, (c_void_p, st_service_p, c_int32, c_enun, c_enun, c_char_p, c_int64, c_enun_p, c_char_p, c_int32)),
    ('MB_SendOneWayMessage', None, (c_void_p, st_service_p, c_int32, c_enun, c_enun, c_char_p, c_int64, c_enun_p, c_char_p, c_int32)),
    ('MB_PeekMessage', st_msg_p, (c_void_p, c_enun, c_enun, c_char_p, c_int32)),
    ('MB_ReplyMessage', c_int32, (c_void_p, st_msg_p, c_int32, c_enun, c_char_p)),
    ('MB_LocateService', st_service_p, (c_void_p, st_service_p, c_char_p, c_int32, c_enun_p, c_char_p, c_int32)),
    ('MB_ReleaseService', c_enun, (st_service_p,)),
    ('MB_Finalize', c_enun, (c_void_p,)),
    ('MB_FinalizeAll', c_enun, ()),
    ('MB_ReleaseMessage', None, (st_msg_p,)),
    ('MB_GetQueueStatus', c_int32, (c_void_p,)),
    ('MB_RegisterService', None, (c_void_p, st_service_p, st_service_p, c_char_p, c_char_p, c_enun_p)),
    ('MB_WaitStart', c_int32, (c_void_p, c_char_p, c_char_p, c_int_p)),
    ('MB_GetTokenName', c_char_p, (c_enun,)),
    ('MB_GetTokenByName', c_enun, (c_char_p,)),
    ('MB_GetTokenByIndex', c_enun, (c_int32,)),
    ('MB_EvtSend', st_msg_p, (c_void_p, st_service_p, c_char_p, c_char_p, c_char_p, c_int, c_int, c_char_p, c_int64, c_enun_p, c_char_p, c_int,)),
    ('MB_EvtSubscribe', c_enun, (c_void_p, st_service_p, c_char_p,)),
    ('MB_EvtUnsubscribe', c_enun, (c_void_p, st_service_p, c_char_p,)),
    # easy message bus API
    ('MB_EasyInitialize', st_easymb_p, (c_char_p, c_enun_p, c_char_p, c_int32)),
    ('MB_EasyRegisterService', None, (st_easymb_p, c_char_p, c_char_p, c_enun_p)),
    ('MB_EasyGetMessage', st_msg_p, (st_easymb_p, c_char_p, c_int32)),
    ('MB_EasySendMessage', st_msg_p, (st_easymb_p, c_char_p, c_int32, c_enun, c_enun, c_void_p, c_int64, c_enun_p, c_char_p, c_int32)),
    ('MB_EasyReplyMessage', c_int32, (st_easymb_p, st_msg_p)),
    ('MB_EasyFinalize', c_enun, (st_easymb_p,)),
    ('MB_EasyWaitStart', c_int32, (c_void_p, c_char_p, c_char_p, c_int_p)),
    ('MB_EasyEvtSend', st_msg_p, (c_void_p, c_char_p, c_char_p, c_char_p, c_int, c_int, c_char_p, c_int64, c_enun_p, c_char_p, c_int,)),
    ('MB_EasyEvtSubscribe', c_enun, (c_void_p, c_char_p,)),
    ('MB_EasyEvtUnsubscribe', c_enun, (c_void_p, c_char_p,)),
)

#
# Private functions
#


def _load_native():
    """ loads native libraries """
    global _load_native, _dll
    _libpath = _find_library("msgbus")
    if not _libpath:
        raise ImportError("Could not find library 'msgbus'.")
    _dll = cdll.LoadLibrary(_libpath)
    _load_native = None


def _init_dll():
    """
    Perform the DLL initialization inside a function, to avoid
    leaving variables in the module scope
    """
    global _init_dll, _dll
    if _load_native:
        _load_native()
    for name, restype, argtypes in _funcdefs:
        func = getattr(_dll, name, None)
        if func is None:
            raise ImportError(
                "Could not find function '%s' on 'msgbus' library." % name)
        func.restype = restype
        func.argtypes = argtypes
    _init_dll = None


def _caller(up=0):
    """Get file name, line number, function name and
       source text of the caller's caller as 4-tuple:
       (file, line, func, text).
       The optional argument 'up' allows retrieval of
       a caller further back up into the call stack.
       Note, the source text may be None and function
       name may be '?' in the returned result.  In
       Python 2.3+ the file name may be an absolute
       path.
    """
    try:  # just get a few frames
        f = traceback.extract_stack(limit=up + 3)
        if f:
            return f[0]
    except:
        pass
    return ('', 0, '', None)

#
# Public APIs
#


def MB_GetTokenName(token):
    """ MB_GetTokenName(token) -> str
    Retrieves the token name for the given token number
    @param token: {int} - token number
    @return: The token name
    """
    global _dll
    if _init_dll:
        _init_dll()
    return _dll.MB_GetTokenName(token)


def MB_GetTokenByName(tokenName):
    """ MB_GetTokenByName(tokenName) -> int
    Retrieves the token number for the given token name
    @param tokenName: {str} - token name
    @return: The token number
    """
    global _dll
    if _init_dll:
        _init_dll()
    return _dll.MB_GetTokenByName(tokenName)


def MB_GetTokenByIndex(tokenIndex):
    """ MB_GetTokenByName(tokenName) -> int
    Retrieves the token number for the given token name
    @param tokenName: {str} - token name
    @return: The token number
    """
    global _dll
    if _init_dll:
        _init_dll()
    return _dll.MB_GetTokenByIndex(tokenIndex)


class MBService:

    """ MBService(address, port, pservice=None, autodel=True) -> instance

        This class represents a message-bus service.
        Instances of this class wraps the result of a call to MB_CreateService.
        @param address: {str} - Host-name or IP address of the service.
        @param port: {int} - Port of the service.
        @param pservice: {st_service_p} - Optional native service pointer to use.
               If this is passed, given address and port are ignored.
        @param autodel: {bool} - Indicates if the native instance should be released with this python object.
        @return: MBService instance
    """

    def __init__(self, address, port, pservice=None, autodel=True):
        global _dll
        if _init_dll:
            _init_dll()
        call = _caller()
        service_name = 0
        if pservice is not None and isinstance(pservice, basestring):
            service_name = pservice
            pservice = None
        self._pservice = _dll.MB_CreateService(
            address, port, service_name, call[0], call[1]) if (pservice is None) else pservice
        service = self._pservice.contents
        self.addr = service.addr
        self.port = service.port
        self._autodel = autodel

    def __del__(self):
        global _dll
        """Object destructor that release the native instance."""
        if self._autodel and _dll:
            _dll.MB_ReleaseService(self._pservice)


class MBMessage:

    """ class MBMessage
        Wraps a native message-bus message.

        Instances of this class should never be directly created by client applications.
    """
    sign = None  # <: signature
    size = None  # <: data size
    token = None  # <: token
    format = None  # <: format
    destname = None  # <: destination service name (optional)
    data = None  # <: data (may be empty)
    trailer = None  # <: trailer
    crc = None  # <: crc32
    # <: remote host name or IP address - for incoming messages only
    remotehost = None

    def __init__(self, pmsg, autodel=True):
        self._pmsg = pmsg
        self._autodel = autodel
        msg = pmsg.contents
        self.sign = msg.sign
        self.size = msg.size
        self.token = msg.token
        self.format = msg.format
        self.destname = msg.destname
        self.data = string_at(msg.data, self.size) if msg.data else ""
        self.trailer = msg.trailer
        self.crc = msg.crc
        self.remotehost = msg.remotehost

    def __del__(self):
        """
        Object destructor that destroy the native instance.
        """
        global _dll
        if self._autodel and _dll:
            _dll.MB_ReleaseMessage(self._pmsg)


class MBContext:

    """ MBContext(my_service, name) -> MBContext

        This class represents a message-bus context. Most available operations
        are executed thru an instance of this class.
        Instances of this class wraps the result of a call to MB_Initialize.

        Please note that when you create a new context, a new native thread is started to listen to messages.
        @param my_service: {MBService} - Service representing the IP address and port number to listen to messages.
        @param name: {str} - Client identfication (used only for debug and logs).
        @return: Context instance.
        @raise MBException: if the call to MB_Initialize results in a status different than SE_SUCCESS.
    """

    def __init__(self, my_service, name):
        global _dll
        if _init_dll:
            _init_dll()
        call = _caller()
        status = c_int32(0)
        self._pctx = _dll.MB_Initialize(
            my_service._pservice, name, c_int32_p(status), call[0], call[1])
        status = status.value
        self._dll = _dll
        if status != 0:
            raise MBException("Error calling MB_Initialize.", status)

    def __del__(self):
        """Object destructor that release the native instance."""
        self.MB_Finalize()

    def MB_GetMessage(self, filtermin=0, filtermax=0):
        """ ctx.MB_GetMessage(filtermin=0, filtermax=0) -> MBMessage

            Retrieves and remove the next message in the message queue (blocks until a
            message is available).
            @param filtermin: {int} - Optional token filter start.
            @param filtermax: {int} - Optional token filter end.
            @return: Message instance, or None if the context has been finalized.
        """
        global _dll
        call = _caller()
        pmsg = _dll.MB_GetMessage(
            self._pctx, filtermin, filtermax, call[0], call[1])
        return MBMessage(pmsg, autodel=False) if pmsg else None

    def MB_PeekMessage(self, filtermin=0, filtermax=0):
        """ ctx.MB_PeekMessage(filtermin=0, filtermax=0) -> MBMessage

            Retrieves the next message in the message queue (don't remove it).
            @param filtermin: {int} - Optional token filter start.
            @param filtermax: {int} - Optional token filter end.
            @return: Message instance, or None if the context has been finalized.
        """
        global _dll
        call = _caller()
        pmsg = _dll.MB_PeekMessage(
            self._pctx, filtermin, filtermax, call[0], call[1])
        return MBMessage(pmsg) if pmsg else None

    def MB_SendMessage(self, destination, token, format=0, data=None, timeout=-1):
        """ ctx.MB_SendMessage(destination, token, format=0, data=None, timeout=-1) -> MBMessage

            Sends a message.
            @param destination: {MBService} - Destination service.
            @param token: {int} - Message token.
            @param format: {int} - Optional message format.
            @param data: {str} - Optional data to send.
            @param timeout: {int} - Optional timeout in microseconds (1000000 = 1s).
            @return: Message instance representing the message reply.
            @raise MBTimeout: if the *timeout* param is set and the message timed-out
            @raise MBException: if the call to native MB_SendMessage results in a status different than SE_SUCCESS
        """
        global _dll
        call = _caller()
        size = len(data) if (data is not None) else 0
        status = c_int32(0)
        pmsg = _dll.MB_SendMessage(self._pctx, destination._pservice, size,
                                   token, format, data, timeout, c_int32_p(status), call[0], call[1])
        status = status.value
        if status != 0:
            if SE_TIMEOUT == status:
                raise MBTimeout("Timed-out on MB_SendMessage.", SE_TIMEOUT)
            raise MBException("Error calling MB_SendMessage.", status)
        return MBMessage(pmsg)

    def MB_SendOneWayMessage(self, destination, token, format=0, data=None, timeout=-1):
        """ ctx.MB_SendOneWayMessage(destination, token, format=0, data=None, timeout=-1)

            Sends a message returning as soon as the message arrives to destination.
            @param destination: {MBService} - Destination service.
            @param token: {int} - Message token.
            @param format: {int} - Optional message format.
            @param data: {str} - Optional data to send.
            @param timeout: {int} - Optional timeout in microseconds (1000000 = 1s).
            @raise MBTimeout: if the *timeout* param is set and the message timed-out
            @raise MBException: if the call to native MB_SendMessage results in a status different than SE_SUCCESS
        """
        global _dll
        call = _caller()
        size = len(data) if (data is not None) else 0
        status = c_int32(0)
        _dll.MB_SendOneWayMessage(self._pctx, destination._pservice, size,
                                  token, format, data, timeout, c_int32_p(status), call[0], call[1])
        status = status.value
        if status != 0:
            if SE_TIMEOUT == status:
                raise MBTimeout(
                    "Timed-out on MB_SendOneWayMessage.", SE_TIMEOUT)
            raise MBException("Error calling MB_SendOneWayMessage.", status)

    def MB_ReplyMessage(self, message, format=FM_UNDEFINED, data=None):
        """ ctx.MB_ReplyMessage(message, format=FM_UNDEFINED, data=None) -> int

            Reply a received message.
            @param message: {MBMessage} - Message to reply (set the response token on this message)
            @param format: {int} - Optional response format.
            @param data: {str} - Optinal response data.
            @return status code
        """
        global _dll
        message._pmsg.contents.token = message.token  # Copy the current token to native message before replying
        size = len(data) if (data is not None) else 0
        status = _dll.MB_ReplyMessage(
            self._pctx, message._pmsg, size, format, data)
        message._autodel = False  # Native message has been deleted already
        return status

    def MB_LocateService(self, hv_service, service_name, maxretries=5):
        """ ctx.MB_LocateService(hv_service, service_name, maxretries=5) -> MBService

            Contacts the hypervisor to locate a service.
            @param hv_service: {MBService} - Service representing the hypervisor location.
            @param service_name: {str} - Name of the service to locate.
            @param maxretries: {int} - Optional number of retries in case of failure.
            @return: Service instance representing the located service, or None if the service has not been found
            @raise MBException: if the call to native MB_LocateService results in a status different than SE_SUCCESS
        """
        global _dll
        call = _caller()
        status = c_int32(0)
        pser = _dll.MB_LocateService(
            self._pctx, hv_service._pservice, service_name, maxretries, c_int32_p(status), call[0], call[1])
        status = status.value
        if status != 0:
            if status == SE_NOTFOUND:
                return None
            raise MBException("Error calling MB_LocateService.", status)
        return MBService(None, None, pser)

    def MB_GetQueueStatus(self):
        """ ctx.MB_GetQueueStatus() -> int

            Retrieves the current queue status.
            @return: status - 0 if the queue is empty, or > 0 if it has any message. (the returned value is not the quantity)
        """
        global _dll
        return _dll.MB_GetQueueStatus(self._pctx)

    def MB_RegisterService(self, my_service, hv_service, service_type, service_name):
        """ ctx.MB_RegisterService(my_service, hv_service, service_type, service_name)

            Register a service on hypervisor. All registered services can be located by others.
            @param my_service: {MBService} - Service representing where to find the given service name.
            @param hv_service: {MBService} - Service representing the hypervisor location.
            @param service_type: {str} - Service type to register.
            @param service_name: {str} - Service name to register.
            @raise MBException: if the call to native MB_RegisterService results in a status different than SE_SUCCESS
        """
        global _dll
        status = c_int32(0)
        _dll.MB_RegisterService(self._pctx, my_service._pservice,
                                hv_service._pservice, service_type, service_name, c_int32_p(status))
        status = status.value
        if status != 0:
            raise MBException("Error calling MB_RegisterService.", status)

    def MB_WaitStart(self, exported_services=None, required_services=None):
        """ ctx.MB_WaitStart(exported_services=None, required_services=None) -> int

            Waits for the hypervisor to allow us really startup.
            This will handle all the startup conversation with the hipervisor and will return non-zero
            if we must terminate for some reason.
            @param exported_services: {str} - Services to export, in the form: "service1:type1|service2:type2|...etc..."
            @param required_services: {str} - List of services that are required, in the form: "service1|service2|..etc..."
        """
        return _dll.MB_WaitStart(self._pctx, exported_services, required_services, None)

    def MB_EvtSend(self, hv_service, subject, type, xml, synchronous=False, sourceid=0, queue=None, timeout=-1):
        """ ctx.MB_EvtSend(hv_service, subject, type, xml, synchronous=False, sourceid=0, queue=None)

            Send a system event.

            @param hv_service: {str} - {MBService} - Service representing the hypervisor location.
            @param subject: {str} - Event subject
            @param type: {str} - Event type
            @param xml: {str} - Event xml data
            @param synchronous: {bool} - Indicates if this is a synchronous event
            @param sourceid: {int} - Optional POSID number that is sent with the event
            @param queue: {str} - Optional event queue
            @param timeout: {int} - Optional timeout in microseconds (1000000 = 1s).
            @raise MBException: if the call to native MB_EvtSend results in a status different than SE_SUCCESS
            @return {MBMessage} - Response message (only makes sense for synchronous events)
        """
        global _dll
        call = _caller()
        sync = 1 if synchronous else 0
        status = c_int(0)
        pmsg = _dll.MB_EvtSend(self._pctx, hv_service._pservice, subject, type,
                               xml, sync, sourceid, queue, timeout, c_int_p(status), call[0], call[1])
        status = status.value
        if status != SE_SUCCESS:
            raise MBException("Error calling MB_EvtSend.", status)
        return MBMessage(pmsg) if pmsg else None

    def MB_EvtSubscribe(self, hv_service, subject):
        """ ctx.MB_EvtSubscribe(hv_service, subject)

            Subscribe to listen a sytem event.

            @param hv_service: {str} - {MBService} - Service representing the hypervisor location.
            @param subject: {str} - Event subject or subjects separated by spaces
            @raise MBException: if the call to native MB_EvtSubscribe results in a status different than SE_SUCCESS
        """
        global _dll
        status = _dll.MB_EvtSubscribe(
            self._pctx, hv_service._pservice, subject)
        if status != SE_SUCCESS:
            raise MBException("Error calling MB_EvtSubscribe.", status)

    def MB_EvtUnsubscribe(self, hv_service, subject):
        """ ctx.MB_EvtUnsubscribe(hv_service, subject)

            Unsubscribe to listen a sytem event.

            @param hv_service: {str} - {MBService} - Service representing the hypervisor location.
            @param subject: {str} - Event subject or subjects separated by spaces
            @raise MBException: if the call to native MB_EvtUnsubscribe results in a status different than SE_SUCCESS
        """
        global _dll
        status = _dll.MB_EvtUnsubscribe(
            self._pctx, hv_service._pservice, subject)
        if status != SE_SUCCESS:
            raise MBException("Error calling MB_EvtUnsubscribe.", status)

    def MB_Finalize(self):
        """ ctx.MB_Finalize()

        Manually finalize and release this context.
        Please note that this method is automatically executed when this context is garbage-collected.
        """
        if self._pctx:
            self._dll.MB_Finalize(self._pctx)
            self._pctx = None

#
# Easy message bus APIs
#


class MBEasyContext(MBContext):

    """ MBEasyContext(name) -> MBEasyContext

        This class represents an "easy" message-bus context. Most available operations
        are executed thru an instance of this class.
        Instances of this class wraps the result of a call to MB_EasyInitialize.

        Please note that when you create a new context, a new native thread is started to listen to messages.
        @param name: {str} - Client identfication (used only for debug and logs).
        @return: Context instance.
        @raise MBException: if the call to MB_EasyInitialize results in a status different than SE_SUCCESS.
    """
    ser_service = None  # <: {MBService} - service representing the local component IP and port
    # <: {MBService} - service representing the hypervisor's IP and port
    hv_service = None

    def __init__(self, name):
        global _dll
        if _init_dll:
            _init_dll()
        call = _caller()
        status = c_int32(0)
        self._dll = _dll
        self._pectx = _dll.MB_EasyInitialize(
            name, c_int32_p(status), call[0], call[1])
        status = status.value
        if status != 0:
            raise MBException("Error calling MB_EasyInitialize.", status)
        ectx = self._pectx.contents
        # Storing the "non-easy" context on "_pctx" allows MBContext methods to
        # work too
        self._pctx = ectx.pctx
        self.ser_service = MBService(None, None, ectx.ser, False)
        self.hv_service = MBService(None, None, ectx.hv, False)

    def __del__(self):
        """Object destructor that release the native instance."""
        self.MB_EasyFinalize()

    def MB_EasyRegisterService(self, service_type, service_name):
        """ ectx.MB_EasyRegisterService(service_type, service_name)

            Register a service on hypervisor. All registered services can be located by others.
            @param service_type: {str} - Service type to register.
            @param service_name: {str} - Service name to register.
            @raise MBException: if the call to native MB_EasyRegisterService results in a status different than SE_SUCCESS
        """
        global _dll
        status = c_int32(0)
        _dll.MB_EasyRegisterService(
            self._pectx, service_type, service_name, c_int32_p(status))
        status = status.value
        if status != 0:
            raise MBException("Error calling MB_EasyRegisterService.", status)

    def MB_EasyGetMessage(self):
        """ ectx.MB_EasyGetMessage() -> MBMessage

            Retrieves and remove the next message in the message queue (blocks until a
            message is available).
            @return: Message instance, or None if the context has been finalized.
        """
        global _dll
        call = _caller()
        pmsg = _dll.MB_EasyGetMessage(self._pectx, call[0], call[1])
        return MBMessage(pmsg, autodel=False) if pmsg else None

    def MB_EasySendMessage(self, dest_name, token, format=0, data=None, timeout=-1):
        """ ectx.MB_EasySendMessage(dest_name, token, format=0, data=None, timeout=-1) -> MBMessage

            Sends a message.
            @param dest_name: {str} - Destination service name.
            @param token: {int} - Message token.
            @param format: {int} - Optional message format.
            @param data: {str} - Optional data to send.
            @param timeout: {int} - Optional timeout in microseconds (1000000 = 1s).
            @return: Message instance representing the message reply.
            @raise MBTimeout: if the *timeout* param is set and the message timed-out
            @raise MBException: if the call to native MB_SendMessage results in a status different than SE_SUCCESS
        """
        global _dll
        call = _caller()
        status = c_int32(0)
        size = len(data) if (data is not None) else 0
        pmsg = _dll.MB_EasySendMessage(
            self._pectx, dest_name, size, token, format, data, timeout, c_int32_p(status), call[0], call[1])
        status = status.value
        if status != 0:
            if SE_TIMEOUT == status:
                raise MBTimeout("Timed-out on MB_EasySendMessage.", SE_TIMEOUT)
            raise MBException("Error calling MB_EasySendMessage.", status)
        return MBMessage(pmsg)

    def MB_EasyReplyMessage(self, message):
        """ ectx.MB_EasyReplyMessage(message) -> int

            Reply a received message.
            @param message: {MBMessage} - Message to reply (set the response token on this message)
            @return status code
        """
        global _dll
        message._pmsg.contents.token = message.token  # Copy the current token to native message before replying
        status = _dll.MB_EasyReplyMessage(self._pectx, message._pmsg)
        message._autodel = False  # Native message has been deleted already
        return status

    def MB_EasyWaitStart(self, exported_services=None, required_services=None):
        """ ectx.MB_EasyWaitStart(exported_services=None, required_services=None) -> int

            Waits for the hypervisor to allow us really startup.
            This will handle all the startup conversation with the hipervisor and will return non-zero
            if we must terminate for some reason.
            @param exported_services: {str} - Services to export, in the form: "service1:type1|service2:type2|...etc..."
            @param required_services: {str} - List of services that are required, in the form: "service1|service2|..etc..."
        """
        global _dll
        return _dll.MB_EasyWaitStart(self._pectx, exported_services, required_services, None)

    def MB_EasyEvtSend(self, subject, type, xml, synchronous=False, sourceid=0, queue=None, timeout=-1):
        """ ctx.MB_EasyEvtSend(subject, type, xml, synchronous=False, sourceid=0, queue=None)

            Send a system event.

            @param subject: {str} - Event subject
            @param type: {str} - Event type
            @param xml: {str} - Event xml data
            @param synchronous: {bool} - Indicates if this is a synchronous event
            @param sourceid: {int} - Optional POSID number that is sent with the event
            @param queue: {str} - Optional event queue
            @param timeout: {int} - Optional timeout in microseconds (1000000 = 1s).
            @raise MBException: if the call to native MB_EasyEvtSend results in a status different than SE_SUCCESS
            @return {MBMessage} - Response message (only makes sense for synchronous events)
        """
        global _dll
        call = _caller()
        sync = 1 if synchronous else 0
        status = c_int(0)
        pmsg = _dll.MB_EasyEvtSend(
            self._pectx, subject, type, xml, sync, sourceid, queue, timeout, c_int_p(status), call[0], call[1])
        status = status.value
        if status != SE_SUCCESS:
            raise MBException("Error calling MB_EasyEvtSend.", status)
        return MBMessage(pmsg) if pmsg else None

    def MB_EasyEvtSubscribe(self, subject):
        """ ctx.MB_EasyEvtSubscribe(subject)

            Subscribe to listen a sytem event.

            @param subject: {str} - Event subject or subjects separated by spaces
            @raise MBException: if the call to native MB_EasyEvtSubscribe results in a status different than SE_SUCCESS
        """
        global _dll
        status = _dll.MB_EasyEvtSubscribe(self._pectx, subject)
        if status != SE_SUCCESS:
            raise MBException("Error calling MB_EasyEvtSubscribe.", status)

    def MB_EasyEvtUnsubscribe(self, subject):
        """ ctx.MB_EasyEvtUnsubscribe(subject)

            Unsubscribe to listen a sytem event.

            @param subject: {str} - Event subject or subjects separated by spaces
            @raise MBException: if the call to native MB_EasyEvtUnsubscribe results in a status different than SE_SUCCESS
        """
        global _dll
        status = _dll.MB_EasyEvtUnsubscribe(self._pectx, subject)
        if status != SE_SUCCESS:
            raise MBException("Error calling MB_EasyEvtUnsubscribe.", status)

    def MB_Finalize(self):
        """ ctx.MB_Finalize()
        Alias for MB_EasyFinalize.
        """
        self.MB_EasyFinalize()

    def MB_EasyFinalize(self):
        """ ctx.MB_EasyFinalize()

        Manually finalize and release this context.
        Please note that this method is automatically executed when this context is garbage-collected.
        """
        if self._pectx:
            self._dll.MB_EasyFinalize(self._pectx)
            self._pectx = None
            self._pctx = None

#
# HTTP Message-bus context
# (Send-only)
#


class MBHttpMessage(object):

    """ Fake MBMessage returned by MBHttpContext"""

    def __init__(self, token, format, data):
        self.token = token
        self.data = data


class MBHttpContext(object):

    """ A fake message-bus context that uses the HTTP interface to send messages (receiving messages is not possible)"""

    def __init__(self, host="127.0.0.1", port=8080, prefix="mwapp", httptimeout=-1):
        self.host = str(host)
        self.port = int(port)
        self.prefix = str(prefix)
        self.httptimeout = httptimeout
        self.httptimeout = socket._GLOBAL_DEFAULT_TIMEOUT if self.httptimeout < 0 else self.httptimeout
        self.conn = None

    def MB_EasyWaitStart(self, *args, **kargs):
        return SE_SUCCESS

    def MB_WaitStart(self, *args, **kargs):
        return SE_SUCCESS

    def MB_Finalize(self, *args, **kargs):
        pass

    def MB_EasyFinalize(self, *args, **kargs):
        pass

    def MB_EasySendMessage(self, dest_name, token, format=FM_PARAM, data=None, timeout=-1):
        """ Sends a message using the HTTP interface """
        data = "" if data is None else base64.b64encode(str(data))
        url = "/%s/services/unknown/%s?token=%s&format=%s&timeout=%d&isBase64=true" % (
            self.prefix, dest_name, token, format, timeout)
        if not self.conn:
            self.conn = httplib.HTTPConnection(
                self.host, self.port, timeout=self.httptimeout)
        try:
            self.conn.request("POST", url, data)
        except:
            # Connection error - reconnect and retry
            self.conn = httplib.HTTPConnection(
                self.host, self.port, timeout=self.httptimeout)
            self.conn.request("POST", url, data)
        res = self.conn.getresponse()
        if int(res.status) == 404:
            raise MBException("Service not found", SE_NOTFOUND)
        if int(res.status) != 200:
            if res.getheader("X-error-code") == "TIMEOUT":
                raise MBTimeout("Timed-out on MB_EasySendMessage.", SE_TIMEOUT)
            raise MBException("HTTP error %s" % res.status, SE_UNDEFINED)
        token, format = map(res.getheader, ('X-token', 'X-format'))
        return MBHttpMessage(int(token or 0), int(format or 0), res.read())

#
# Exceptions
#


class MBException(Exception):

    """ Exception raised when any message bus error happen.

        All message bus exceptions contain a field called 'errorcode'
        that can be used to check the native error code that generated
        the exception.
    """

    def __init__(self, message, errorcode):
        message += " Error code: %d(%s)." % (errorcode, sys_errget(errorcode))
        Exception.__init__(self, message)
        self.errorcode = errorcode


class MBTimeout(MBException):

    """Exception raised to signalize a timeout."""
    pass



# -*- coding: utf-8 -*-
#
# Python wrapper for the "systools" library
#

# Python standard modules
import sys
import os
import threading
import traceback
import ctypes
import ctypes.util
from datetime import datetime
from cStringIO import StringIO
from ctypes import c_char_p, c_int32, cdll, create_string_buffer
try:
    from ctypes import windll
except:
    windll = None

import pickle
from contextlib import contextmanager
from time import time

# Exported (public) symbols and functions

__all__ = ("sys_log", "sys_log_trace", "sys_log_debug", "sys_log_info",
           "sys_log_warning", "sys_log_error", "sys_log_fatal",
           "sys_log_exception", "sys_parsefilepath", "sys_errget",
           "sys_errset",)

__pid = os.getpid()
__compname = os.environ.get("HVCOMPNAME") or "unknown"


def _find_library(name):
    """ finds a native library in the system, given it's name """
    platform = "win32" if sys.platform.startswith("win") else "linux" if (
        sys.platform.find("linux") != -1) else "macos"
    prefix, suffix, envvar = ({"win32": ("", ".dll", "PATH"), "linux": (
        "lib", ".so", "LD_LIBRARY_PATH"), "macos": ("lib", ".dylib", "DYLD_LIBRARY_PATH")})[platform]
    filename = "%s%s%s" % (prefix, name, suffix)
    for libpath in (os.path.join(dirname, filename) for dirname in ["./"] + os.environ.get(envvar, "").split(os.path.pathsep)):
        if os.path.exists(libpath):
            return libpath
    libpath = ctypes.util.find_library(name)
    if libpath and os.path.exists(libpath):
        return libpath
    return None


def sys_log(level, message):
    # The date/time is formatted as "ISO 8601", but with only three digits for
    # the "milliseconds"
    strtime = datetime.now().isoformat()[:23]
    if len(strtime) == 19:
        strtime += ".000"  # isoformat() not always generate the milliseconds
    # Current thread id
    threadid = threading.currentThread().ident or 0
    # Caller identification
    file, line, func, text = __caller(1)
    file = file[max(file.rfind('\\'), file.rfind('/')) + 1:]
    # Print each line individually
    try:
        str_msg = str(message)
    except UnicodeEncodeError:
        str_msg = str(message.encode('utf-8'))
    for logline in str_msg.splitlines():
        print "<%s> %s - %s; [%s - %s - %s:%d (%d:0x%X)]" % (level, strtime, logline, __compname, func, file, line, __pid, threadid)
    # Flushs the standard output stream
    try:
        sys.stdout.flush()
    except IOError:
        pass  # This may happen if the parent process is dead - ignore


def sys_log_trace(message):
    sys_log("TRACE", message)


def sys_log_debug(message):
    sys_log("DEBUG", message)


def sys_log_info(message):
    sys_log("INFO ", message)


def sys_log_warning(message):
    sys_log("WARN ", message)


def sys_log_error(message):
    sys_log("ERROR", message)


def sys_log_fatal(message):
    sys_log("FATAL", message)


def sys_log_exception(message, level="ERROR"):
    buf = StringIO()
    traceback.print_exc(file=buf)
    buf = buf.getvalue()
    if isinstance(buf, unicode):
        buf = buf.encode("UTF-8")
    if isinstance(message, unicode):
        message = message.encode("UTF-8")
    sys_log(level, "%s\n%s" % (message, buf))


def sys_parsefilepath(path):
    """ sys_parsefilepath(path) -> str
    Parses a file path, replacing environment variables inside braces: '{ENV}'
    @param path: {str} - File path with optional environment variable. E.g.: "{MY_PATH}/subfolder"
    @return {str} - the file path with the environment variable replaced. E.g.: "/tmp/subfolder"
    """
    start, end = (path.find("{"), path.find("}"))
    if (start != -1 and end != -1 and end > start + 1):
        envvar = os.environ.get(path[start + 1:end])
        if envvar is not None:
            return path[:start] + envvar + path[end + 1:]
    return path


def sys_errget(code):
    """
    sys_errget(code) -> the error description for the given error code

    @brief - retrieve the error description for the given error code

    @param - sysErrors code - code to retrieve the message
    @return the error description for the given error code
    """
    try:
        return Syslib().errget(code)
    except:
        return "(could not get error message)"


def sys_errset(code, msg):
    """
    sys_errset(code,msg) -> SE_SUCCESS for success setting the message or SE_NOTFOUND if the given code wasn't found

    @brief - sets a new error message for an existent error code

    @param - sysErrors code - error code to set the message
    @param - const char *message - message to set (this message must remain allocated in the caller memory stack)
    @return - SE_SUCCESS for success setting the message or SE_NOTFOUND if the given code wasn't found
    """
    return Syslib().errset(code, msg)

#
# DLL Function definitions
#


class Syslib(object):
    _dll__spam = None
    _dynmsgs__spam = list()
    _funcdefs__spam = (
        ('sys_errget', c_char_p, (c_int32,)),
        ('sys_errset', c_int32, (c_int32, c_char_p,)),
    )

    def __init__(self):
        """ starts the systools.dll library  """

        if self._dll__spam:
            return

        _libpath = _find_library("systools")
        # verify if path is valid
        if not _libpath:
            raise ImportError("Could not find library 'systools'.")
        # retrieve the correct loader based on platform
        dllLoader = windll if (sys.platform[:3] == 'win') else cdll
        # load the library
        self._dll__spam = dllLoader.LoadLibrary(_libpath)
        # load all needed functions
        for name, restype, argtypes in self._funcdefs__spam:
            func = getattr(self._dll__spam, name, None)
            if func is None:
                raise ImportError(
                    "Could not find function '%s' on 'systools' library." % name)
            func.restype = restype
            func.argtypes = argtypes

    def errget(self, code):
        """ gets error description """
        return self._dll__spam.sys_errget(code)

    def errset(self, code, msg):
        """ gets error description """
        lmsg = create_string_buffer(msg)
        self._dynmsgs__spam.append(lmsg)
        return self._dll__spam.sys_errset(code, lmsg)


def __caller(up=0):
    """Get file name, line number, function name and
       source text of the caller's caller as 4-tuple:
       (file, line, func, text).
    """
    try:  # just get a few frames
        f = traceback.extract_stack(limit=up + 3)
        if f:
            return f[0]
    except:
        pass
    return ('', 0, '', None)


def debug2file(data, append=False):
    '''Write to a temp file some data. Default file is /tmp/debug2file.txt
    data: Some str data
    append: append to end of file (append=True) or overwrite file

    TODO: Don't be a slackware, as me, use tempifile from Python stdlib
    '''

    mode = 'a' if append else 'w'
    with open('/tmp/debug2file.txt', mode) as dbg:
        dbg.write(str(data))


def dump2pickle(fp, data, mode='wb'):
    '''Dump objects to a pickle file that can be loaded in an external shell
    Supported types: dict, list, tuple, int, str, float
    fp = file name with full path to pickle file.
    data = data to dump
    mode = write once (default) or append

    Example:

    employees = get_employees_list()
    dump_to_pickle('/tmp/employees', employees, mode='wb')

    On console:

    >>> from pickle import load
    >>> emp = load('/tmp/employees')
    Voilá!
    '''
    if type(data) in [dict, list, tuple, int, str, float]:
        with open(fp, mode) as handle:
            pickle.dump(data, handle)


@contextmanager
def timeit(context=None):
    '''Context manager to measure slices of code
    context: Text to identify the tracker

    Example:

    with timeit('measuring a simple "for" statement'):
        c = 0
        for i in range(100):
            c += 1
    '''
    start_time = time()
    try:
        yield
    finally:
        args = context, str(time() - start_time)
        msg = '#### Context: %s. That slice of code took %s seconds ####' % args
        sys_log_debug(msg)



# -*- coding: utf-8 -*-
# Module name: cfgtools.py
# Module Description: Python wrapper for the cfgtools API
#
# Copyright (C) 2012 MWneo Corporation
#
# $Id$
# $Revision$
# $Date$
# $Author$

"""
Python wrapper for the cfgtools API.
Minimal usage example:
>>> import cfgtools
>>> cfg = cfgtools.read("/path/to/my.cfg")
>>> some_value = cfg.find_value("Group1.Subgroup.SomeValue")
>>> # .. do something with some_value ..
>>> # Iterate over all "level 1" groups and their keys
>>> for group in cfg.groups:
        for key in group.keys:
           print "%s.%s = %s"%(group.name,key.name,key.value())
>>> print "done!"
"""

import re
import os
import xml.sax
import xml.sax.handler
import xml.sax.saxutils as saxutils

# Key types
ARRAY = 0x00000001
STRING = 0x00000002
DATA = 0x00000004
INTEGER = 0x00000008
REAL = 0x00000010
DATE = 0x00000020
BOOLEAN = 0x00000040

_TYPES_TAGMAP = {"array": ARRAY, "string": STRING, "data": DATA, "integer": INTEGER, "real": REAL, "date": DATE, "boolean": BOOLEAN}
_TYPES_MAP = {ARRAY: "array", STRING: "string", DATA: "data", INTEGER: "integer", REAL: "real", DATE: "date", BOOLEAN: "boolean"}
_RE_PATH = re.compile(r"[\.|/\\]")  # allowed separators: [. | / \]


def read(file_or_path):
    """ cfgtools.read("file_or_path") -> Configuration
    Parses a configuration file and creates a new Configuration instance.
    @param file_or_path - file instance or path to a file
    @return Configuration instance
    @raise SAXException if the configuration could not be parsed
    """
    handler = _CfgSaxHandler()
    xml.sax.parse(file_or_path, handler)
    cfg = handler.cfg
    bundle = os.environ.get('BUNDLEDIR')
    loader = os.environ.get('LOADERCFG')
    override = os.environ.get('CFGOVERRIDE')
    override = (str(override).lower() == "yes")
    override = (override and bundle is not None and loader is not None)
    if override:
        try:
            loader = loader[len(bundle):][:-4]
            bundle = bundle.split('/')
            bundle = bundle[len(bundle) - 1] or bundle[len(bundle) - 2]
        except:
            override = False
    if override:
        def _group_visit(gs, v):
            for g in gs.groups:
                _group_visit(g, v + '_' + g.name)
            for k in gs.keys:
                ovar = v + '_' + k.name
                oval = os.environ.get(ovar)
                if oval is not None:
                    k.values = [oval]
        for grp in cfg.groups:
            _group_visit(grp, bundle + '_' + loader + '_' + grp.name)
    return cfg


def read_string(string):
    """ cfgtools.read_string(string) -> Configuration
    Parses a configuration file from a string and creates a new Configuration instance.
    @param string - string with the XML data
    @return Configuration instance
    @raise SAXException if the configuration could not be parsed
    """
    handler = _CfgSaxHandler()
    xml.sax.parseString(string, handler)
    return handler.cfg


class Configuration(object):
    __slots__ = ('groups', 'version', )

    def __init__(self):
        self.groups = []
        self.version = "1.0"

    def group(self, name):
        """obj.group(name) -> Group
        Finds the first child group with the given name.
        @param name - the name of the group
        @return a Group instance, or None if not found
        """
        return _byname(self.groups, name)

    def find_group(self, path):
        """obj.find_group(path) -> Group
        Finds a group in the hierarchy, following the given path
        @param path - the path to look for, where groups are separated
               using one of: [. | / \]. E.g.: "MyGroup.SubGroup.OtherGroup"
        @return a Group instance, or None if not found
        """
        return _bypath(self, path)

    def find_key(self, path):
        """obj.find_key(path) -> Key
        Finds a key in the hierarchy, following the given path
        @param path - the path to look for, where groups/keys are separated
               using one of: [. | / \]. E.g.: "MyGroup.SubGroup.MyKey"
        @return a Key instance, or None if not found
        """
        return _bypath(self, path, iskey=True)

    def find_value(self, path, default=None):
        """obj.find_value(path) -> str
        Finds a key value in the hierarchy, following the given path
        @param path - the path to look for, where groups/keys are separated
               using one of: [. | / \]. E.g.: "MyGroup.SubGroup.MyKey"
        @param default - the default value to return if the given path
                         is not found
        @return the value found, or default
        """
        key = _bypath(self, path, iskey=True)
        return default if key is None else key.value()

    def find_values(self, path):
        """obj.find_values(path) -> list
        Finds a key values list in the hierarchy, following the given path
        @param path - the path to look for, where groups/keys are separated
               using one of: [. | / \]. E.g.: "MyGroup.SubGroup.MyKey"
        @return the values list found, or None if not found
        """
        key = _bypath(self, path, iskey=True)
        return None if key is None else key.values

    def add_group(self, name, descr=None):
        """obj.add_group(self, "name", descr=None) -> Group
        Creates a new group as a child of this element.
        @param name - group name
        @param descr - optional group description
        """
        grp = Group(name, descr, self)
        if self.groups:
            self.groups[len(self.groups) - 1].next = grp
        self.groups.append(grp)
        return grp

    def write(self, output_io):
        """obj.write(output_io)
        Serializes this configuration structure XML to an IO.
        @param output_io - IO to write XML data to
        """
        self._write(output_io, 0)

    def _write(self, out, level):
        out.write("""<?xml version="1.0" encoding="UTF-8"?>\n<config version="%s">\n""" % self.version)
        for group in self.groups:
            group._write(out, level + 1)
        out.write("""</config>\n""")


class Group(object):
    __slots__ = ('name', 'descr', 'keys', 'groups', 'next', 'parent')

    def __init__(self, name=None, descr=None, parent=None):
        self.name = name
        self.descr = descr
        self.keys = []
        self.groups = []
        self.next = None
        self.parent = parent

    def group(self, name):
        """obj.group(name) -> Group
        Finds the first child group with the given name.
        @param name - the name of the group
        @return a Group instance, or None if not found
        """
        return _byname(self.groups, name)

    def key(self, name):
        """obj.key(name) -> Key
        Finds a child key with the given name.
        @param name - the name of the key
        @return a Key instance, or None if not found
        """
        return _byname(self.keys, name)

    def key_value(self, keyname, default=None):
        """obj.key_value(keyname, default=None) -> str
        Retrieves the value of a child key with the given name.
        @param keyname - the name of the key
        @param default - default value to return if the key is not found
        @return {str} - the key value, or None if not found
        """
        key = _byname(self.keys, keyname)
        return default if key is None else key.value()

    def find_group(self, path):
        """obj.find_group(path) -> Group
        Finds a group in the hierarchy, following the given path
        @param path - the path to look for, where groups are separated
               using one of: [. | / \]. E.g.: "MyGroup.SubGroup.OtherGroup"
        @return a Group instance, or None if not found
        """
        return _bypath(self, path)

    def find_key(self, path):
        """obj.find_key(path) -> Key
        Finds a key in the hierarchy, following the given path
        @param path - the path to look for, where groups/keys are separated
               using one of: [. | / \]. E.g.: "MyGroup.SubGroup.MyKey"
        @return a Key instance, or None if not found
        """
        return _bypath(self, path, iskey=True)

    def find_value(self, path, default=None):
        """obj.find_value(path) -> str
        Finds a key value in the hierarchy, following the given path
        @param path - the path to look for, where groups/keys are separated
               using one of: [. | / \]. E.g.: "MyGroup.SubGroup.MyKey"
        @param default - the default value to return if the given path
                         is not found
        @return the value found, or default
        """
        key = _bypath(self, path, iskey=True)
        return default if key is None else key.value()

    def find_values(self, path):
        """obj.find_values(path) -> list
        Finds a key values list in the hierarchy, following the given path
        @param path - the path to look for, where groups/keys are separated
               using one of: [. | / \]. E.g.: "MyGroup.SubGroup.MyKey"
        @return the values list found, or None if not found
        """
        key = _bypath(self, path, iskey=True)
        return None if key is None else key.values

    def add_group(self, name, descr=None):
        """obj.add_group(self, "name", descr=None) -> Group
        Creates a new group as a child of this element.
        @param name - group name
        @param descr - optional group description
        """
        grp = Group(name, descr, self)
        if self.groups:
            self.groups[len(self.groups) - 1].next = grp
        self.groups.append(grp)
        return grp

    def add_key(self, name, descr=None):
        """obj.add_key(self, "name", descr=None) -> Key
        Creates a new key as a child of this element.
        @param name - key name
        @param descr - optional key description
        """
        key = Key(name, descr, self)
        if self.keys:
            self.keys[len(self.keys) - 1].next = key
        self.keys.append(key)
        return key

    def _write(self, out, level):
        out.write('%s<group' % ("\t" * level))
        if self.name is not None:
            out.write(' name=%s' % saxutils.quoteattr(self.name.encode("UTF-8")))
        if self.descr is not None:
            out.write(' descr=%s' % saxutils.quoteattr(self.descr.encode("UTF-8")))
        out.write('>\n')
        for group in self.groups:
            group._write(out, level + 1)
        for key in self.keys:
            key._write(out, level + 1)
        out.write('%s</group>\n' % ("\t" * level))


class Key(object):
    __slots__ = ('name', 'descr', 'type', 'values', 'next', 'parent')

    def __init__(self, name=None, descr=None, parent=None):
        self.name = name
        self.descr = descr
        self.type = None
        self.values = []
        self.next = None
        self.parent = parent

    def value(self):
        """obj.value() -> str
        Gets the first value of this key.
        @return the value or None if this key is empty
        """
        return self.values[0] if self.values else None

    def add_value(self, value):
        """obj.add_value("value")
        Adds a value to this key
        @param value - value to add
        @return self for convenience
        """
        self.values.append(value)
        return self

    def _write(self, out, level):
        out.write('%s<key' % ("\t" * level))
        if self.name is not None:
            out.write(' name=%s' % saxutils.quoteattr(self.name.encode("UTF-8")))
        if self.descr is not None:
            out.write(' descr=%s' % saxutils.quoteattr(self.descr.encode("UTF-8")))
        out.write('>\n')
        if self.values:
            level += 1
            typestr = _TYPES_MAP[self.type]
            if self.type == ARRAY:
                typestr = "string"
                out.write('%s<array>\n' % ("\t" * level))
                level += 1
            for val in self.values:
                out.write('%s<%s>%s</%s>\n' % ("\t" * level, typestr, saxutils.escape(val.encode("UTF-8")), typestr))
            if self.type == ARRAY:
                level -= 1
                out.write('%s</array>\n' % ("\t" * level))
            level -= 1
        out.write('%s</key>\n' % ("\t" * level))


def _bypath(node, path, index=0, iskey=False):
    "private - finds keys or groups recursivelly given a path"
    if node is not None:
        if isinstance(path, basestring):
            path = _RE_PATH.split(path)
        next_idx = index + 1
        if index < len(path):
            if iskey and (next_idx == len(path)):
                return node.key(path[index])
            return _bypath(node.group(path[index]), path, next_idx, iskey)
        return node
    return None


def _byname(array, name):
    "private - finds a node in a list"
    for obj in array:
        if obj.name == name:
            return obj
    return None


class _CfgSaxHandler(xml.sax.handler.ContentHandler):
    "private - SAX handler used to parse configuration files"
    def __init__(self):
        self.cfg = Configuration()
        self.current = self.cfg
        self.currval = None

    def startElement(self, name, attrs):
        if name == "group":
            if isinstance(self.current, (Configuration, Group)):
                self.current = self.current.add_group(name=attrs.get("name"), descr=attrs.get("descr"))
        elif name == "key":
            if isinstance(self.current, Group):
                self.current = self.current.add_key(name=attrs.get("name"), descr=attrs.get("descr"))
        elif name in _TYPES_TAGMAP:
            if self.current.type is None:
                self.current.type = _TYPES_TAGMAP[name]
            if self.currval is None and name != "array":
                self.currval = u""
        elif name == "config":
            if self.current == self.cfg:
                ver = attrs.get("version")
                if ver:
                    self.current.version = ver

    def characters(self, content):
        if self.currval is not None and isinstance(self.current, Key):
            self.currval += content

    def endElement(self, name):
        if self.currval is not None and isinstance(self.current, Key):
            self.current.values.append(self.currval)
            self.currval = None
        elif name == "group" or name == "key":
            self.current = self.current.parent



# -*- coding: utf-8 -*-
# Module name: persistence.py
# Module Description: Python wrapper for the persistence layer
#
# Copyright © 2008 MWneo Corporation
#
# $Id$
# $Revision$
# $Date$
# $Author$

"""
Python wrapper for the persistence layer
Minimal usage example:
>>> import persistence
>>> driver = persistence.Driver()
>>> conn = driver.open(mbcontext) # You need a message-bus context here
>>> cursor = conn.select("SELECT * FROM mydb.MyTable")
>>> print "   |   ".join(cursor.get_names())
>>> for row in cursor:
>>>     print "|".join([str(entry) for entry in row])
>>> print "END"
"""

import ctypes
import ctypes.util
import os.path
import sys
from ctypes import c_int, c_size_t, c_void_p, c_char_p, byref, POINTER, create_string_buffer, cdll
try:
    from ctypes import windll
except:
    windll = None

_dll = None
_udll = None


def _find_library(name):
    """ finds a native library in the system, given it's name """
    platform = "win32" if sys.platform.startswith("win") else "linux" if (
        sys.platform.find("linux") != -1) else "macos"
    prefix, suffix, envvar = ({"win32": ("", ".dll", "PATH"), "linux": (
        "lib", ".so", "LD_LIBRARY_PATH"), "macos": ("lib", ".dylib", "DYLD_LIBRARY_PATH")})[platform]
    filename = "%s%s%s" % (prefix, name, suffix)
    for libpath in (os.path.join(dirname, filename) for dirname in ["./"] + os.environ.get(envvar, "").split(os.path.pathsep)):
        if os.path.exists(libpath):
            return libpath
    libpath = ctypes.util.find_library(name)
    if libpath and os.path.exists(libpath):
        return libpath
    return None


# Native APR types


_apr_status_t = c_int
_apr_size_t = c_size_t
_apr_pool_t = c_void_p
_apr_pool_tp = POINTER(_apr_pool_t)
_apr_dbd_driver_t = c_void_p
_apr_dbd_driver_tp = POINTER(_apr_dbd_driver_t)
_apr_dbd_t = c_void_p
_apr_dbd_tp = POINTER(_apr_dbd_t)
_apr_dbd_results_t = c_void_p
_apr_dbd_results_tp = POINTER(_apr_dbd_results_t)
_apr_dbd_row_t = c_void_p
_apr_dbd_row_tp = POINTER(_apr_dbd_row_t)
_apr_dbd_transaction_t = c_void_p
_apr_dbd_transaction_tp = POINTER(_apr_dbd_transaction_t)
_apr_dbd_prepared_t = c_void_p
_apr_dbd_prepared_tp = POINTER(_apr_dbd_prepared_t)

_APR_EOF_OPTIONS = (-1, 70014)
_APR_EGENERAL = 20014

# Other types

c_int_p = POINTER(c_int)
c_char_pp = POINTER(c_char_p)

#
# DLL Function definitions
#

_funcdefs = (  # For libapr-1
    ('apr_pool_create_ex', _apr_status_t, (_apr_pool_tp, _apr_pool_t, c_void_p, c_void_p,)),
    ('apr_pool_destroy', None, (_apr_pool_t,)),
    ('apr_initialize', _apr_status_t, ()),
    ('apr_pool_initialize', _apr_status_t, ()),
    ('apr_strerror', c_char_p, (_apr_status_t, c_char_p, _apr_size_t,)),
    ('apr_pool_userdata_set', _apr_status_t, (c_char_p, c_char_p, c_void_p, _apr_pool_t)),
    ('apr_pool_userdata_get', _apr_status_t, (c_char_pp, c_char_p, _apr_pool_t)),
)

_ufuncdefs = (  # For libaprutil-1
    ('apr_dbd_init', _apr_status_t, (_apr_pool_t,)),
    ('apr_dbd_get_driver', _apr_status_t, (_apr_pool_t, c_char_p, _apr_dbd_driver_tp,)),
    ('apr_dbd_open', _apr_status_t, (_apr_dbd_driver_t, _apr_pool_t, c_void_p, _apr_dbd_tp,)),
    ('apr_dbd_set_dbname', _apr_status_t, (_apr_dbd_driver_t, _apr_pool_t, _apr_dbd_t, c_char_p)),
    ('apr_dbd_close', _apr_status_t, (_apr_dbd_driver_t, _apr_dbd_t,)),
    ('apr_dbd_query', c_int, (_apr_dbd_driver_t, _apr_dbd_t, c_int_p, c_char_p,)),
    ('apr_dbd_select', c_int, (_apr_dbd_driver_t, _apr_pool_t, _apr_dbd_t, _apr_dbd_results_tp, c_char_p, c_int,)),
    ('apr_dbd_num_cols', c_int, (_apr_dbd_driver_t, _apr_dbd_results_t,)),
    ('apr_dbd_num_tuples', c_int, (_apr_dbd_driver_t, _apr_dbd_results_t,)),
    ('apr_dbd_get_row', c_int, (_apr_dbd_driver_t, _apr_pool_t, _apr_dbd_results_t, _apr_dbd_row_tp, c_int,)),
    ('apr_dbd_get_entry', c_char_p, (_apr_dbd_driver_t, _apr_dbd_row_t, c_int,)),
    ('apr_dbd_get_name', c_char_p, (_apr_dbd_driver_t, _apr_dbd_results_t, c_int,)),
    ('apr_dbd_error', c_char_p, (_apr_dbd_driver_t, _apr_dbd_t, c_int,)),
    ('apr_dbd_escape', c_char_p, (_apr_dbd_driver_t, _apr_pool_t, c_char_p, _apr_dbd_t,)),
    ('apr_dbd_transaction_start', c_int, (_apr_dbd_driver_t, _apr_pool_t, _apr_dbd_t, _apr_dbd_transaction_tp,)),
    ('apr_dbd_transaction_end', c_int, (_apr_dbd_driver_t, _apr_pool_t, _apr_dbd_transaction_t,)),
    ('apr_dbd_transaction_mode_get', c_int, (_apr_dbd_driver_t, _apr_dbd_transaction_t,)),
    ('apr_dbd_transaction_mode_set', c_int, (_apr_dbd_driver_t, _apr_dbd_transaction_t, c_int)),
    ('apr_dbd_prepare', c_int, (_apr_dbd_driver_t, _apr_pool_t, _apr_dbd_t, c_char_p, c_char_p, _apr_dbd_prepared_tp)),
    ('apr_dbd_pselect', c_int, (_apr_dbd_driver_t, _apr_pool_t, _apr_dbd_t, _apr_dbd_results_tp, _apr_dbd_prepared_t, c_int, c_int, c_char_pp)),
    ('apr_dbd_pquery', c_int, (_apr_dbd_driver_t, _apr_pool_t, _apr_dbd_t, c_int_p, _apr_dbd_prepared_t, c_int, c_char_pp)),
)

#
# Private functions
#


def _aprcall_err(fname, errfunc, *args):
    global _dll, _udll
    func = getattr(_udll, fname, False) or getattr(_dll, fname)
    status = func(*args)
    if status != 0:
        raise AprDbdException(errfunc(status), status)


def _aprcall(fname, *args):
    global _dll, _udll
    """Calls an apr function that returns a status code and raises an error if the status is non-zero"""
    func = getattr(_udll, fname, False) or getattr(_dll, fname)
    status = func(*args)
    if status != 0:
        raise AprException("Error on function '%s'" % (fname), status)


def _apr_strerror(statcode):
    global _dll
    """Call apr_strerror to retrieve a human-readable message for an error code"""
    buf = create_string_buffer("", 512)
    return _dll.apr_strerror(statcode, buf, 512)


def _pool_create():
    """Creates an APR memory pool"""
    pool = _apr_pool_t(0)
    _aprcall('apr_pool_create_ex', byref(pool), None, None, None)
    return pool


def _pool_destroy(pool):
    global _dll
    """Destroys an APR memory pool"""
    _dll.apr_pool_destroy(pool)


def _params_to_charpp(params):
    """Converts a dictionary of parameters into a (char **) array"""
    if not params:
        params = {}
    # ctypes array types are constructed by multiplying a type by the number
    # of elements
    array_type = (c_char_p * (len(params) + 1))
    array = array_type()
    index = 0
    for name, value in params.iteritems():
        if value is None:
            value = "null"
        array[index] = ("%s=%s" % (name, value)).encode("UTF-8")
        index += 1
    return array


def _load_native():
    """ loads native libraries """
    global _load_native, _udll, _dll
    dllLoader = windll if (sys.platform[:3] == 'win') else cdll
    # libaprutil-1 (I don't know exactly why, but on MacOS X we must load "aprutil" BEFORE "apr"
    # otherwise it will fail to locate a symbol)
    _libpath = _find_library("aprutil-1") or _find_library("libaprutil-1")
    if not _libpath:
        raise ImportError("Could not find library 'libaprutil-1'.")
    _udll = dllLoader.LoadLibrary(_libpath)

    # libapr-1
    _libpath = _find_library("apr-1") or _find_library("libapr-1")
    if not _libpath:
        raise ImportError("Could not find library 'libapr-1'.")
    _dll = dllLoader.LoadLibrary(_libpath)
    _load_native = None


def _init_dll():
    """
    Perform the DLL initialization inside a function, to avoid
    leaving variables in the module scope
    """
    global _init_dll, _dll, _udll
    if _load_native:
        _load_native()
    defs = (('libaprutil-1', _udll, _ufuncdefs), ('libapr-1', _dll, _funcdefs))
    for libname, dll, funcdefs in defs:
        for name, restype, argtypes in funcdefs:
            func = getattr(dll, name, None)
            if func is None:
                raise ImportError(
                    "Could not find function '%s' on '%s' library." % (name, libname))
            func.restype = restype
            func.argtypes = argtypes
    # Call initialization functions
    _aprcall('apr_initialize')
    _aprcall('apr_pool_initialize')
    _aprcall('apr_dbd_init', _pool_create())
    _init_dll = None


#
# Public APIs
#


__all__ = ('Driver', 'Connection', 'Cursor', 'Row', 'DbdException', 'AprException', 'AprDbdException',
           'APR_DBD_TRANSACTION_COMMIT', 'APR_DBD_TRANSACTION_ROLLBACK', 'APR_DBD_TRANSACTION_IGNORE_ERRORS',)

APR_DBD_TRANSACTION_COMMIT = 0x00  # commit the transaction
APR_DBD_TRANSACTION_ROLLBACK = 0x01  # rollback the transaction
APR_DBD_TRANSACTION_IGNORE_ERRORS = 0x02  # ignore transaction errors


class Driver(object):

    """ class Driver(object)
    This is the class that should be instantiated to use the persistence module.
    Instances of this class encapsulates an "apr dbd driver".
    """

    def __init__(self, name="persistcomp"):
        """ Driver(name="persistcomp") -> instance
        Creates a new persistence driver
        @raise AprException on any APR error
        """
        if _init_dll:
            _init_dll()
        self._pool = _pool_create()
        self._driver = _apr_dbd_driver_t()
        _aprcall('apr_dbd_get_driver', self._pool, name, byref(self._driver))

    def __del__(self):
        """private - object destructor that releases the native instance"""
        if _pool_destroy:  # This will be None if the module has already been unloaded
            _pool_destroy(self._pool)

    def open(self, mbcontext, dbname=None, service_name=None):
        """ obj.open(self, mbcontext) -> Connection
        Opens a new database connection.
        @param mbcontext: {msgbus.MBContext} - Message-bus context to use, or {str} (string connection)
        @param dbname: {str} - Optional dbname to set right after opening the connection (the same as calling "set_dbname")
        @param service_name: {str} - Optional persistence service name to connect to
        @return {Connection} - the connection instance
        @raise AprException on any APR error
        """
        if hasattr(mbcontext, '_pctx'):
            conn_string = mbcontext._pctx
            mbctxt = mbcontext
        else:
            conn_string = str(mbcontext)
            mbctxt = None
        return Connection(conn_string, self, dbname, service_name, mbctxt)


class Connection(object):

    """ class Connection(object)
    Represents a database connection.
    This class is instantiated by #Driver.open
    """
    _open = False   # Indicates if this connection is open

    def __init__(self, conn_string, driverobj, dbname=None, service_name=None, mbcontext=None):
        """private - Connection constructor"""
        self._driverobj = driverobj  # This avoids the driver being unloaded while the connection is active
        self._driver = driverobj._driver
        self._pool = _pool_create()
        self._handle = _apr_dbd_t()
        self._trans = None
        self._service_name = service_name
        self._number_of_instances = -1
        self._instaces = []
        if self._service_name:
            _aprcall('apr_pool_userdata_set', self._service_name,
                     "PERSISTENCE_SERVICE", None, self._pool)
        _aprcall('apr_dbd_open', self._driver, self._pool,
                 conn_string, byref(self._handle))
        self._open = True
        if dbname and self.is_multi_instance():
            self.set_dbname(dbname)
        if mbcontext is not None:
            from os import environ
            from msgbus import MBService, TK_PERSIST_GETINSTNBR, TK_PERSIST_INITINPROGRESS, FM_PARAM, TK_SYS_ACK
            while True:
                msg = mbcontext.MB_SendMessage(mbcontext.MB_LocateService(MBService(environ["HVIP"], int(environ["HVPORT"])), (service_name or "Persistence")), TK_PERSIST_GETINSTNBR, FM_PARAM, "", -1)
                if msg.token == TK_PERSIST_INITINPROGRESS:
                    from time import sleep
                    sleep(1)
                    continue
                elif msg.token == TK_SYS_ACK:
                    params = msg.data.split("\0")
                    if len(params) > 0:
                        self._number_of_instances = int(params[0])
                    if len(params) > 1:
                        self._instaces = [int(p) for p in params[1:]]
                break

    def __del__(self):
        """private - object destructor that releases the native instance"""
        if _pool_destroy:  # This will be None if the module has already been unloaded
            self.close()

    def _error_message(self, errnum):
        """private - retrieves an error message using 'apr_dbd_error' """
        global _udll
        return _udll.apr_dbd_error(self._driver, self._handle, errnum)

    def close(self):
        """ obj.close()
        Closes this connection (it cannot be used anymore after this).
        @raise AprException on any APR error
        """
        if self._open:
            if self._trans:
                self.transaction_end()
            _aprcall('apr_dbd_close', self._driver, self._handle)
            _pool_destroy(self._pool)
            self._open = False
            del self._driverobj

    def get_number_of_instances(self):
        """ obj.get_number_of_instances() -> int
        Retrieves the number of databases instances available in the Persistence component is applicable
        @return Number of instances in the connected persistence component
        """
        return self._number_of_instances

    def get_instances(self):
        """ obj.get_instances() -> int[]
        Retrieves the instance ids list to be used in the set_dbname call.
        @return List of int with valid instance ids
        """
        return self._instaces

    def is_multi_instance(self):
        """ obj.is_multi_instance() -> bool
        Checks the current working mode of the persistence component.
        @return True if the current working mode is multi-instance, othwerwise False.
        """
        global _udll
        if not self._open:
            raise DbdException("Connection already closed")
        status = _udll.apr_dbd_set_dbname(
            self._driver, self._pool, self._handle, None)
        return (status != _APR_EGENERAL)

    def set_dbname(self, dbname):
        """ obj.set_dbname(dbname)
        @param dbname The database instance which the user desires to work.
        @raise AprException if an invalid instance was passed in the dbname parameter.
        """
        if not self._open:
            raise DbdException("Connection already closed")
        _aprcall('apr_dbd_set_dbname', self._driver,
                 self._pool, self._handle, dbname)

    def query(self, sql):
        """ obj.query(sql) -> int
        Executes a SQL statement and returns the number of rows affected.
        @param sql: {str} - sql statement. multiple statements are allowed by separating
                    them with '\0' characters. The last (or single) statement should be
                    terminated with a '\0', but that is optional since this method ensures
                    this (if you need better performance, add it yourself)
        @return {int} - the number of rows affected
        @raise AprException on any APR error
        @raise DbdException if this connection is closed
        """
        if not self._open:
            raise DbdException("Connection already closed")
        nrows = c_int(0)
        if not sql.endswith("\0"):
                # The sql buffer MUST end with two '\0' (one is already added
                # by ctypes, but we better ensure)
            sql += "\0\0"
        _aprcall_err('apr_dbd_query', self._error_message,
                     self._driver, self._handle, byref(nrows), sql)
        return nrows.value

    def pquery(self, procname, **kargs):
        """ obj.pquery(procname, **kargs) -> int
        Executes a prepared procedure and returns the number of affected rows
        @param procname: {str} - the prepared procedure name
        @param kargs: each keyword argument is a procedure parameter
        @return {int} - the number of rows affected
        @raise AprException on any APR error
        @raise DbdException if this connection is closed
        """
        if not self._open:
            raise DbdException("Connection already closed")
        nrows = c_int(0)
        pool = _pool_create()
        try:
            procedure = _apr_dbd_prepared_t()
            # Prepare the procedure
            _aprcall_err('apr_dbd_prepare', self._error_message,
                         self._driver, pool, self._handle, "", procname, byref(procedure))
            # Build the list of parameters
            array = _params_to_charpp(kargs)
            _aprcall_err('apr_dbd_pquery', self._error_message, self._driver,
                         pool, self._handle, byref(nrows), procedure, len(kargs), array)
        finally:
            _pool_destroy(pool)
        return nrows.value

    def select(self, sql):
        """ obj.select(sql) -> Cursor
        Executes a SQL "select" query and returns a cursor to the results
        @param sql: {str} - sql statement
        @return {Cursor} - a cursor with the query results
        @raise AprException on any APR error
        @raise DbdException if this connection is closed
        """
        if not self._open:
            raise DbdException("Connection already closed")
        return Cursor(self, sql)

    def pselect(self, procname, **kargs):
        """ obj.pselect(procname, **kargs) -> Cursor
        Executes a prepared procedure query and returns a cursor to the results
        @param procname: {str} - the prepared procedure name
        @param kargs: each keyword argument is a procedure parameter
        @return {ProcedureCursor} - a cursor with the query results
        @raise AprException on any APR error
        @raise DbdException if this connection is closed
        """
        if not self._open:
            raise DbdException("Connection already closed")
        return ProcedureCursor(self, procname, kargs)

    def escape(self, string):
        """obj.escape(string) -> str
        Escapes a string to be inserted into an SQL estatement.
        @param string: {str} - the string to be escaped
        @return {str} - the escaped string
        """
        global _udll
        if isinstance(string, unicode):
            string = string.encode("UTF-8")
        return _udll.apr_dbd_escape(self._driver, self._pool, string, self._handle)

    def transaction_start(self):
        """obj.transaction_start()
        Starts a transaction in the database
        @raise AprException on any APR error
        @raise DbdException if this connection is closed
        """
        if not self._open:
            raise DbdException("Connection already closed")
        if self._trans:
            return
        trans = _apr_dbd_transaction_t()
        _aprcall_err('apr_dbd_transaction_start', self._error_message,
                     self._driver, self._pool, self._handle, byref(trans))
        self._trans = trans

    def transaction_end(self):
        """obj.transaction_end()
        Closes the currently open transaction in the database
        @raise AprException on any APR error
        @raise DbdException if this connection is closed
        """
        if not self._open:
            raise DbdException("Connection already closed")
        if not self._trans:
            return
        _aprcall_err('apr_dbd_transaction_end', self._error_message,
                     self._driver, self._pool, self._trans)
        self._trans = None

    def transaction_mode_get(self):
        """obj.transaction_mode_get()
        Retrieves the current transaction mode (APR_DBD_TRANSACTION_COMMIT,
        APR_DBD_TRANSACTION_ROLLBACK or APR_DBD_TRANSACTION_IGNORE_ERRORS)
        @raise DbdException if this connection is closed
        """
        global _udll
        if not self._open:
            raise DbdException("Connection already closed")
        if not self._trans:
            return
        return _udll.apr_dbd_transaction_mode_get(self._driver, self._trans)

    def transaction_mode_set(self, mode):
        """obj.transaction_mode_get()
        Sets the current transaction mode
        @param mode: {int} - APR_DBD_TRANSACTION_COMMIT, APR_DBD_TRANSACTION_ROLLBACK or APR_DBD_TRANSACTION_IGNORE_ERRORS
        @raise DbdException if this connection is closed
        """
        global _udll
        if not self._open:
            raise DbdException("Connection already closed")
        if not self._trans:
            return
        _udll.apr_dbd_transaction_mode_set(
            self._driver, self._trans, int(mode))


class Cursor(object):

    """ class Cursor(object)
    Cursor that holds a result set for a SQL query.
    This class is instantiated by #Connection.select
    """

    def __init__(self, connobj, sql):
        """private - Cursor constructor"""
        global _udll
        self._connobj = connobj  # This avoids the connection being unloaded while the cursor is active
        self._driver = connobj._driver
        self._handle = connobj._handle
        self._pool = _pool_create()
        self._res = _apr_dbd_results_t()
        _aprcall_err('apr_dbd_select', connobj._error_message,
                     self._driver, self._pool, self._handle, byref(self._res), sql, 0)
        self._maxcols = _udll.apr_dbd_num_cols(self._driver, self._res)
        self._colnames = []
        self._coltypes = []
        for i in xrange(self._maxcols):
            self._colnames.append(self.get_name(i))
            value = c_char_p()
            _aprcall('apr_pool_userdata_get', byref(value), "COLTYPE", self._pool)
            self._coltypes.append(value.value)

    def __del__(self):
        """private - object destructor that releases the native instance"""
        if _pool_destroy:  # This will be None if the module has already been unloaded
            _pool_destroy(self._pool)
        del self._connobj

    def __iter__(self):
        """creates an iterator for the rows of this cursor"""
        return _RowsIterator(self)

    def cols(self):
        """ obj.cols() -> int
        Retrieves the total number of columns on this cursor
        @return {int} - number of columns
        """
        return self._maxcols

    def rows(self):
        """ obj.rows() -> int
        Retrieves the total number of rows on this cursor
        @return {int} - number of rows
        """
        global _udll
        return _udll.apr_dbd_num_tuples(self._driver, self._res)

    def get_name(self, col):
        """ obj.get_name(col) -> str
        Retrieves the name of a column
        @param col: {int} - index of the column (zero-based)
        @return {str} - Name of the column, or None if it does not exist
        """
        global _udll
        return _udll.apr_dbd_get_name(self._driver, self._res, col)

    def get_type(self, col):
        """ obj.get_name(col) -> str
        Retrieves the name of a column
        @param col: {int} - index of the column (zero-based)
        @return {str} - Type of column description
        """
        return self._coltypes[col] if col < len(self._coltypes) else "SQL_UNKNOWN"

    def get_names(self):
        """ obj.get_names() -> list
        Retrieves the name of all columns
        @return {tuple} - Tuple of column names
        """
        return tuple(self._colnames)

    def get_row(self, rownum):
        """ obj.get_row() -> Row
        Retrieves a single row of this cursor
        @param rownum: {int} - index of the row (zero-based)
        @return {Row} - the Row instance, or None if that row does not exist
        """
        try:
            return Row(self, rownum)
        except AprException, ex:
            if ex.errorcode in _APR_EOF_OPTIONS:  # No such row
                return None
            raise


class ProcedureCursor(Cursor):

    """ class ProcedureCursor(Cursor)
    Cursor that holds a result set for a procedure query.
    This class is instantiated by #Connection.pselect
    """

    def __init__(self, connobj, procname, params):
        """private - Cursor constructor"""
        global _udll
        self._connobj = connobj  # This avoids the connection being unloaded while the cursor is active
        self._driver = connobj._driver
        self._handle = connobj._handle
        self._pool = _pool_create()
        self._res = _apr_dbd_results_t()
        procedure = _apr_dbd_prepared_t()
        # Prepare the procedure
        _aprcall_err('apr_dbd_prepare', connobj._error_message, self._driver,
                     self._pool, self._handle, "", procname, byref(procedure))
        # Build the list of parameters
        array = _params_to_charpp(params)
        _aprcall_err('apr_dbd_pselect', connobj._error_message, self._driver,
                     self._pool, self._handle, byref(self._res), procedure, 0, len(params), array)
        self._maxcols = _udll.apr_dbd_num_cols(self._driver, self._res)
        self._colnames = []
        self._coltypes = []
        for i in xrange(self._maxcols):
            self._colnames.append(self.get_name(i))
            value = c_char_p()
            _aprcall('apr_pool_userdata_get', byref(value), "COLTYPE", self._pool)
            self._coltypes.append(value.value)

    def get_output(self, name):
        value = c_char_p()
        _aprcall('apr_pool_userdata_get', byref(value), name, self._pool)
        result = None
        try:
            result = value.value
        except:
            pass
        return result


class Row(object):

    """ class Row(object)
    Represents a row in a cursor.
    This class is instantiated by #Cursor.get_row, or by iterating on a Cursor instance
    """

    def __init__(self, cursorobj, rownum):
        """private - Row constructor"""
        self._cursorobj = cursorobj  # This avoids the cursor being unloaded while the row is active
        self._res = cursorobj._res
        self._pool = cursorobj._pool
        self._driver = cursorobj._driver
        self._row = _apr_dbd_row_t()
        _aprcall('apr_dbd_get_row', self._driver, self._pool,
                 self._res, byref(self._row), rownum)

    def __iter__(self):
        """creates an iterator for the entries of this row"""
        return _EntriesIterator(self, self._cursorobj.cols())

    def get_entry(self, col):
        """ obj.get_entry(col) -> str
        Retrieves a single entry of this row.
        @param col: {int} - column index of the entry (zero-based), or
                    {str} - column name of the entry
        @return {str} - the row entry, or None if it is NULL on the database
        """
        global _udll
        if isinstance(col, str):
            col = self._cursorobj._colnames.index(col)
        return _udll.apr_dbd_get_entry(self._driver, self._row, col)


class _RowsIterator(object):

    """private - rows iterator"""

    def __init__(self, cursor):
        self.idx = 0
        self.cur = cursor

    def __iter__(self):
        return self

    def next(self):
        val = self.cur.get_row(self.idx)
        if val is None:
            raise StopIteration
        self.idx += 1
        return val


class _EntriesIterator(object):

    """private - entries iterator"""

    def __init__(self, row, maxcols):
        self.idx = 0
        self.row = row
        self.max = maxcols

    def __iter__(self):
        return self

    def next(self):
        if self.idx >= self.max:
            raise StopIteration
        val = self.row.get_entry(self.idx)
        self.idx += 1
        return val


class DbdException(Exception):

    """Base exception for this module (all other exceptions descend from this)"""
    pass


class AprException(DbdException):

    """Exception raised for APR errors"""

    def __init__(self, message, errorcode):
        message += (". Status: (%d). APR Message: %s" %
                    (errorcode, _apr_strerror(errorcode)))
        self.errorcode = errorcode
        DbdException.__init__(self, message)


class AprDbdException(AprException):

    """Exception raised for APR-DBD errors"""

    def __init__(self, message, errorcode):
        self.errorcode = errorcode
        DbdException.__init__(self, message)
